# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Implements an emission-line profile fitting class.

.. warning::

    Although it may still be possible to use :class:`Elric`, it is
    currently not actively used by the survey-level operation of the
    MaNGA DAP. See instead :class:`mangadap.proc.sasuke.Sasuke`.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""

import os
import time
import logging

import numpy
from scipy import interpolate, integrate, optimize
from scipy.special import erf

import astropy.constants
from astropy.modeling.polynomial import Legendre1D

from ..par.parset import KeywordParSet
from ..par.emissionlinedb import EmissionLineDB
from ..util.fileio import init_record_array
from ..util.log import log_output
from ..util.pixelmask import SpectralPixelMask
from ..util import lineprofiles
from .spatiallybinnedspectra import SpatiallyBinnedSpectra
from .stellarcontinuummodel import StellarContinuumModel
from .spectralfitting import EmissionLineFit
from .util import sample_growth

from matplotlib import pyplot

# Add strict versioning
# from distutils.version import StrictVersion

# Main fitting engine --------------------------------------------------
class LineProfileFit:
    r"""
    Simultaneously fit multiple line profiles. Currently only allows
    one to fit using an
    :class:`~mangadap.util.lineprofiles.NCompLineProfile` object. The
    fitting algorithm used is `scipy.optimize.least_squares`_ with
    fitting method 'trf' to allow for bounds.

    .. todo::
        For multicomponent lines, set the first normalization to be the
        normalization for the sum of all components, then force the
        normalization of the subcomponents to be ordered from highest to
        lowest and bounded from 0 to 1.

    Args:
        x (array-like):
            1D vector with the independent variable
        y (array-like):
            1D vector with the dependent variable
        profile_list (:obj:`list`, :class:`~mangadap.util.lineprofiles.NCompLineProfile`):
            The profile(s) to fit to the dependent variable.
        base_order (int): (**Optional**) The order of the Legendre
            polynomial to include in the model for the baseline trend in
            y below the fitted line profile(s).
        error (array-like, optional):
            1D vector with the error in the dependent variables. If
            not provided, no error weighting is performed during the
            fitting process, and the covariance *will not* be
            constructed.
        mask (array-like, optional):
            1D boolean array used to ignore values in `y` during the
            fit.
        par (array-like, optional):
            1D vector with the initial guess for model parameters.
            The number of parameters much match the expectation based
            on the provided list of profiles and the order of the
            baseline polynomial. If not provided, the parameters are
            initialized to 0.
        fixed (array-like, optional):
            1D vector with flags used to fix parameters during the
            fit. The number of parameters much match the expectation
            based on the provided list of profiles and the order of
            the baseline polynomial. If not provided, all parameters
            are freely fit.
        bounds (:obj:`tuple`): (**Optional**) 2-tuple with two
            array-like elements giving the upper and lower bound for
            each parameter. The length of each array element must
            match the number of parameters. For an unbounded problem,
            set ``bounds=None``, or use numpy.inf with an appropriate
            sign to disable bounds on all or some variables.
        run_fit (bool): (**Optional**) Flag to run the fit upon
            instantiation of the object, which defaults to True.  If set
            to False, the object is initialized but the fit is not
            executed, and can be executed later using :func:`fit`.
        construct_covariance (bool): (**Optional**) Flag to construct
            the covariance matrix based on the result object provided by
            `scipy.optimize.least_squares`_, which defaults to True.  If
            set to False, the covariance matrix is set to None.
        verbose (int): (**Optional**) Verbosity level for
            `scipy.optimize.least_squares`_; default is 0.

    Attributes:
        x (`numpy.ndarray`_): Independent variable of length :math:`M`.
        y (`numpy.ndarray`_): Dependent variable to be fit of length
            :math:`M`.
        err (`numpy.ndarray`_): Error, :math:`\sigma`, in the dependent
            variable of length :math:`M`.
        mask (`numpy.ndarray`_): Flag to fit dependent variables of length
            :math:`M`.
        nlines (int): Number of lines being fit.
        base_order (int): The order of the Legendre polynomial include
            in the model; see `astropy.modeling.polynomial.Legendre1D`_.
        model (`astropy.modeling.models.CompoundModel`_): The compound
            model being fit composed of the line profiles and the
            baseline (if a baseline is being fit).  That is, this
            defines the function :math:`f(\mathbf{x}|\mathbf{\theta})`,
            where :math:`\mathbf{x}` is the dependent variable and
            :math:`\mathbf{\theta}` is the list of variables.  The
            fitting algorithm minimizes the (error-weighted) residuals
            to approximate :math:`y=f(\mathbf{x}|\mathbf{\theta})`.
        npar (int): Total number of parameters in the model.  The is the
            number of parameters per line and the number of parameters
            included in the baseline.
        nfitpar (int): Number of *free* parameters, defined as
            :math:`N`.
        par (`numpy.ndarray`_): Full list of model parameters, including
            those parameters that have been fixed.
        fixed (`numpy.ndarray`_): Flags to fit (False) or fix (True) a
            given parameter.
        bounds (:obj:`tuple`): 2-tuple with two array-like
            elements giving the upper and lower bound for each
            parameter.  For an unbounded problem, this is set to
            ``bounds=(-numpy.inf,numpy.inf)``.
        result (`scipy.optimize.OptimizeResult`_): Object with the
            results from `scipy.optimize.least_squares`_.  The
            best-fitting parameters, :math:`\mathbf{\theta}`, is
            returned as ``result.x``.
        cov (`numpy.ndarray`_): The formal covariance matrix for the fit.
            The `scipy.optimize.OptimizeResult`_ object provides the
            Jacobian of the model, an :math:`M \times N` array with
            elements

            .. math::

                J_{ij} = \left.\frac{\partial f_i}{\partial
                \theta_j}\right|_{\mathbf{\theta}}

            at location in parameter space of the best-fitting model.
            This is used to construct the covariance matrix by taking
            the inverse of the curvature matrix:

            .. math::

                \mathbf{\alpha}_{kl} = \left[\frac{1}{\sigma_{i}}
                J_{ik}\right]^{\rm T} \left[\frac{1}{\sigma_{i}}
                J_{il}\right].

            That is, :math:`\mathbf{C} = \mathbf{\alpha}^{-1}`.

    Raises:
        TypeError:
            Raised if the provided profile objects are not instances
            of :class:`~mangadap.util.lineprofiles.NCompLineProfile`.
        ValueError:
            Raised if any of the provided parameter arrays (par,
            fixed) are not one-dimensional or the number of
            parameters is not as expected based on the number of
            profile and baseline parameters.

    .. todo::
        - Provide a better initialization for the parameters.
        - Need to provide the best-fitting parameters and errors as full
          vectors, including the fixed parameters.

    """
    def __init__(self, x, y, profile_list, base_order=None, error=None, mask=None, par=None,
                 fixed=None, bounds=None, run_fit=True, construct_covariance=True, verbose=0):

        # Check the input list of profiles
        _profile_list = profile_list if isinstance(profile_list, list) else [profile_list]
        self.nlines = len(_profile_list)
        for i in range(self.nlines):
            if not isinstance(_profile_list[i], lineprofiles.NCompLineProfile):
                raise TypeError('LineProfileFit only works with NCompLineProfile objects.')

        # Construct and check the parameter vectors
        self.npar = numpy.sum([ p.npar for p in _profile_list])
        self.base_order = -1 if base_order is None else base_order
        if self.base_order > -1:
            self.npar += (self.base_order + 1)
        if par is not None and len(par.shape) != 1:
            raise ValueError('Input parameter array should only be one-dimensional.')
        if fixed is not None and fixed.shape != par.shape:
            raise ValueError('Input parameter and fixed flags should have the same shape.')
        _par = None if par is None else par.ravel()
        if _par is not None and len(_par) != self.npar:
            raise ValueError('Incorrect number of parameters provided')
        self.par = numpy.zeros(self.npar, dtype=float) \
                            if _par is None else _par.astype(float)
        self.fixed = numpy.zeros(self.npar, dtype=bool) \
                            if fixed is None else fixed.astype(bool)
        if not isinstance(bounds, tuple):
            raise ValueError('Bounds must have type tuple.')
        if len(bounds) != 2:
            raise ValueError('Bounds must only have two elements.')
        if len(bounds[0]) != self.npar or len(bounds[1]) != self.npar:
            raise ValueError('Lower and/or upper bounds have an incorrect number of parameters.')
        self.bounded = bounds is not None
        self.bounds = (-numpy.inf, numpy.inf) if bounds is None else bounds

        # Get the number of free parameters
        self.nfitpar = numpy.sum(numpy.invert(self.fixed))

        # Construct the model
        self.model = _profile_list[0].profile
        for i in range(1,self.nlines):
            self.model += _profile_list[i].profile
        if self.base_order > -1:
            self.model += Legendre1D(self.base_order)
        # Fix the parameters
        if numpy.sum(self.fixed) > 0:
            self.model.fixed[ self.model.param_names[fixed] ] = True

#        print(self.model)

        # Need to setup how to tie fluxes, velocities, dispersions
        self.x = None
        self.y = None
        self.err = None
        self.mask = None
        self.result = None
        self.cov = None

        if run_fit:
            self.fit(x, y, error=error, mask=mask, construct_covariance=construct_covariance,
                     verbose=verbose)


    def _assign_par(self, par):
        """
        Fixed parameters are ignored!
        """
        if len(par) not in [ self.npar, self.nfitpar ]:
            raise ValueError('Incorrect number of parameters provided')
        self.par[numpy.invert(self.fixed)] = par[numpy.invert(self.fixed)] \
                                                    if len(par) == self.npar else par
        # Then tie parameters...

        
    def _chi(self, par):
        return self._resid(par)/self.err


    def _resid(self, par):
        self._assign_par(par)
        ymodel = self._quick_sample(self.x)
        return (self.y-ymodel)


    def _calculate_covariance_matrix(self, verbose=0):
        if self.result is None or self.err is None:
            if self.result is not None:
                warnings.warn('Cannot calculate covariance matrix without error vector!')
            return None
#        invsig = numpy.array([numpy.ma.power(self.err,-1.)]*self.result.jac.shape[1]).T
#        a = (invsig.ravel()*self.result.jac.ravel()).reshape(self.result.jac.shape)
#        a = numpy.dot(a.T,a)
        a = numpy.dot(self.result.jac.T,self.result.jac)
        if verbose > 0:
            print('Curvature:')
            print('    ', a)
        try:
            return numpy.linalg.inv(a)
        except:
            return numpy.linalg.pinv(a)


    def _quick_sample(self, x):
        """
        Sample without providing/checking new input parameters.
        """
        #print(*self.par.ravel())
        return self.model.evaluate(x, *self.par.ravel())


    def sample(self, x, par=None):
        if par is not None:
            self._assign_par(par)
        return self._quick_sample(x)

    
    def fit(self, x, y, error=None, mask=None, construct_covariance=True, verbose=0):
        """
        Fit the line profiles provided upon initialization to the data
        using `scipy.optimize.least_squares`_.
        """
        # Check that the lengths of all the arrays match
        if len(x.shape) != 1:
            raise ValueError('Can only fit single spectra!')
        if x.shape != y.shape \
                or (error is not None and error.shape != y.shape) \
                or (mask is not None and mask.shape != y.shape):
            raise ValueError('All vector shapes must be identical.')

        # If the input is a masked array, combine the two masks; the
        # first line below works for both normal numpy.ndarrays and
        # MaskedArray
        _mask = numpy.ma.getmaskarray(y)
        if mask is not None:
            _mask |= mask

        # Set the internal vectors to be numpy.ndarrays that only include the unmasked values
        indx = numpy.invert(_mask)
        self.x = numpy.array(x[indx])
        self.y = numpy.array(y[indx])
        self.err = None if error is None else numpy.array(error[indx])

        if self.y.size <= self.npar:
            raise ValueError('Insufficient data points ({0}) to fit with this function ' \
                             '(npar={1})!'.format(self.y.size, self.npar))

#        print('Initial guess parameters inside LineProfileFit:')
#        print(self.par)
#        print(self.bounds)

        try:
            self.result = optimize.least_squares(self._resid if self.err is None else self._chi,
                                                 self.par, bounds=self.bounds, jac='2-point',
                                                 method='trf', loss='linear', verbose=verbose)
        except ValueError as e:
            print('ValueError: {0}'.format(e))
            print(self.par)
            print(self.bounds)
            self.result = optimize.OptimizeResult(success=False)
            return

#        try:
#            self.result = optimize.least_squares(self._resid if self.err is None else self._chi,
#                                                 self.par, bounds=self.bounds, jac='2-point',
#                                                 #method='lm', loss='linear', verbose=1)
#                                                 method='trf', loss='linear', verbose=2)
#        except:
#            exc_info = sys.exc_info()
#            warnings.warn('Error during fit: {0}:: {1}'.format(exc_info[0], exc_info[1]))
#            self.result = OptimizeResult(x=self.par, cost=None, fun=None, jac=None, grad=None,
#                                         optimality=None, active_mask=None, nfev=None, njev=None,
#                                         status=None, success=False)
#            return

        self.cov = None
        if construct_covariance:
            self.cov = self._calculate_covariance_matrix(verbose=verbose)

        if verbose > 0:
            print('Result: ')
            print('    ', self.result.x)
            if self.cov is not None:
                print('Covariance: ')
                print('    ', self.cov)


class ElricPar(KeywordParSet):
    """
    Elric emission-line fitting parameters.

    The defined parameters are:

    .. include:: ../tables/elricpar.rst
    """
    def __init__(self, emission_lines=None, base_order=None, window_buffer=None,
                 guess_redshift=None, guess_dispersion=None, minimum_snr=None, pixelmask=None,
                 stellar_continuum=None):

        arr_in_fl = [ numpy.ndarray, list, int, float ]
        in_fl = [ int, float ]

        pars =     [ 'emission_lines', 'base_order', 'window_buffer', 'guess_redshift',
                     'guess_dispersion', 'minimum_snr', 'pixelmask', 'stellar_continuum' ]
        values =   [ emission_lines, base_order, window_buffer, guess_redshift, guess_dispersion,
                     minimum_snr, pixelmask, stellar_continuum ]
        defaults = [ None, -1,   25.0, None, None, 0.0, None, None ]
        dtypes =   [ EmissionLineDB, int, in_fl, arr_in_fl, arr_in_fl, in_fl, SpectralPixelMask,
                     StellarContinuumModel ]

        super(ElricPar, self).__init__(pars, values=values, defaults=defaults, dtypes=dtypes)


    def toheader(self, hdr):
        hdr['LPBASEO'] = (self['base_order'], 'Baseline Legendre polynomial order')
        hdr['LPWIN'] = (self['window_buffer'], 'Buffer for fitting window (ang)')
        return hdr


    def fromheader(self, hdr):
        self['base_order'] = hdr['LPBASEO']
        self['window_buffer'] = hdr['LPWIN']



class TiedLineProfile:
    def __init__(self, profile, base_profile, relative_flux=None, mean_separation=None,
                 relative_stddev=None):
        self.profile = profile
        self.base_profile = base_profile
        self.relative_flux = relative_flux
        self.mean_separation = mean_separation
        self.relative_stddev = relative_stddev


    def tie_flux(self):
        self.profile.set_flux(self.relative_flux*self.base_profile.flux())


    def tie_mean(self):
        self.profile.set_mean(self.base_profile.moment(order=1)+self.velocity_separation)


    def tie_stddev(self):
        self.profile.set_stddev(self.relative_stddev*self.base_profile.moment(order=2))


class ElricFittingWindow:
    r"""

    A utility class for the fitting windows used by Elric

    The class can be instantiated as fully None.

    Args:
        nlines (int): (**Optional**) The number of lines to be fit
            simultaneously.
        db_indx (`numpy.ndarray`_): (**Optional**) An array with the
            database index (0-based) of each line to be fit.
        line_index (`numpy.ndarray`_): (**Optional**) An array with the
            index numbers, read from the database, of each line to be
            fit.
        restwave (`numpy.ndarray`_): (**Optional**) An array with the rest
            wavelengths of each line to be fit.
        profile_set (`numpy.ndarray`_): (**Optional**) An array with the
            profile objects that define the functional form of each
            line.
        fixed_par (`numpy.ndarray`_): (**Optional**) An array with the
            fixed parameters for *all* parameters in the model to be
            fit.
        bounds (`numpy.ndarray`_): (**Optional**) A two-column array with
            the lower (first column) and upper (second column) bounds
            for the fit parameters.
        log_bounds (`numpy.ndarray`_): (**Optional**) The range of the
            boundary should be considered as logarithmic when testing if
            a parameter is near its boundary.
        output_model (bool): (**Optional**) Include the best-fitting
            model in the composite emission-line model for each
            spectrum.  This is only flagged as true if ALL the emission
            lines in the fitting window are to be included according to
            the emission-line database.
        tied_pairs (`numpy.ndarray`_): (**Optional**) The series of
            tied profiles (:class:`TiedLineProfile` objects) that are
            used to tie parameters of the model.
        tied_funcs (`numpy.ndarray`_): (**Optional**) The member functions
            of the :class:`TiedLineProfile` objects that should be
            called *sequentially* to tie model parameters.
    """
    def __init__(self, nlines=None, db_indx=None, line_index=None, restwave=None, profile_set=None,
                 fixed_par=None, bounds=None, log_bounds=None, output_model=None, tied_pairs=None,
                 tied_funcs=None):

        self.nlines = nlines
        self.db_indx = db_indx
        self.line_index = line_index
        self.restwave = restwave
        self.profile_set = profile_set
        self.init_par = numpy.array([ p.par.copy().ravel() for p in self.profile_set ])
        self.fixed_par = fixed_par
        self.bounds = bounds
        self.log_bounds = log_bounds
        self.output_model = output_model
        self.tied_pairs = tied_pairs
        self.tied_funcs = tied_funcs


    def append(self, db_indx, line_index, restwave, profile, fixed_par, bounds, log_bounds,
               output_model):
        """
        Append another line to the fitting window.  Must be a single line!
        """

        # Check that input is for a single line
        if not isinstance(db_indx, int) or not isinstance(line_index, (int, numpy.integer)):
            raise TypeError('Appended index must be a single integer.')
        if not isinstance(restwave, float):
            raise TypeError('Appended restwavelength must be a single float.')
        if not isinstance(profile, lineprofiles.NCompLineProfile):
            raise TypeError('Appended profile must have type NCompLineProfile.')
        if len(fixed_par.shape) != 1 or len(fixed_par) != profile.npar:
            raise TypeError('Appended fixed parameter must be a vector with the correct length.')
        if len(bounds.shape) != 2 or bounds.shape[0] != profile.npar or bounds.shape[1] != 2:
            raise TypeError('Appended bounds must have shape: {0}'.format(tuple(profile.npar,2)))
        if len(log_bounds.shape) != 1 or len(log_bounds) != profile.npar:
            raise TypeError('Appended log_bounds parameter must be a vector of the correct length.')
        if not isinstance(output_model, bool):
            raise TypeError('Flag to output model must be boolean.')
        
        self.nlines += 1
        self.db_indx = numpy.append(self.db_indx, db_indx)
        self.line_index = numpy.append(self.line_index, line_index)
        self.restwave = numpy.append(self.restwave, restwave)
        self.profile_set = numpy.append(self.profile_set, profile)
        self.init_par = numpy.append(self.init_par, profile.par.copy(), axis=0)
        self.fixed_par = numpy.append(self.fixed_par, fixed_par.astype(bool).reshape(1,-1), axis=0)
        self.bounds = numpy.append(self.bounds, bounds.reshape(1,-1,2), axis=0)
        self.log_bounds = numpy.append(self.log_bounds, log_bounds.astype(bool).reshape(1,-1),
                                       axis=0)
        self.output_model &= output_model


    def reset_init_mean(self, indx, newmean):
        self.profile_set[indx].set_mean(newmean)
        self.init_par[indx,:] = self.profile_set[indx].par


    def reinit_profiles(self):
        for i in range(self.nlines):
            self.profile_set[i].assign_par(self.init_par[i,:])


class Elric(EmissionLineFit):
    """
    ELRIC: Emission-Line Regression and Inference Class

    https://en.wikipedia.org/wiki/Edward_Elric

    Use LineProfileFit to fit the emission-line properties in a set of
    spectra.

    .. todo::
        - Implement some scheme to penalize multicomponent fits at low S/N

    """
    def __init__(self, bitmask, wave=None, flux=None, emission_lines=None, error=None, mask=None,
                 stellar_continuum=None, base_order=-1, window_buffer=25, guess_redshift=None,
                 guess_dispersion=None, default_dispersion=20.0, run_fit=False, loggers=None,
                 quiet=False):

        EmissionLineFit.__init__(self, 'elric', bitmask)
        # Attributes kept by SpectralFitting:
        #   fit_type='emission_line', bitmask=bitmask, par=None
        # Attributes kept by EmissionLineFit:
        #   fit_method='elric'

        # Declare the attributes kept for convenience
        self.default_dispersion = default_dispersion

        self.loggers=None
        self.quiet=None

        self.nspec = None
        self.nwave = None
        self.emission_lines = None
        self.neml = None

        self.nwindows = None
        self.fitting_window = None

        self.wave = None
        self.sres = None
        self.flux = None
        self.fluxnc = None
        self.error = None
        self.mask = None
        self.continuum = None
        self.no_continuum = None

        self.base_order = None
        self.window_buffer = None
        self.redshift = None
        self.dispersion = None

        self.bestfit = None

        if not run_fit:
            return

        if wave is None or flux is None or emission_lines is None:
            raise ValueError('Cannot run fit without wavelength and flux vectors, and ' \
                             'emission-line database.')

        self.fit(wave, flux, emission_lines, error=error, mask=mask,
                 stellar_continuum=stellar_continuum, base_order=base_order,
                 window_buffer=window_buffer, guess_redshift=guess_redshift,
                 guess_dispersion=guess_dispersion, loggers=loggers, quiet=quiet)


    # TODO: Convert this to a DataTable
    @staticmethod
    def _per_fitting_window_dtype(nwin, max_npar, mask_dtype):
        r"""
        Construct the record array data type for the output fits
        extension.
        """

        return [ ('BINID',int),
                 ('BINID_INDEX',int),
                 ('MASK', mask_dtype, (nwin,)),
                 ('NPIXFIT',int,(nwin,)),
                 ('PAR',float,(nwin,max_npar)),
                 ('ERR',float,(nwin,max_npar)),
                 ('LOBND',float,(nwin,max_npar)),
                 ('UPBND',float,(nwin,max_npar)),
                 ('FIXED',bool,(nwin,max_npar)),
#                 ('TIED',int,(nwin,max_npar)),
                 ('IGNORE',bool,(nwin,max_npar)),
                 ('CHI2',float,(nwin,)),
                 ('RCHI2',float,(nwin,)),
                 ('RMS',float,(nwin,)),
                 ('RESID',float,(nwin,7)),
                 ('FRAC_RMS',float,(nwin,)),
                 ('FRAC_RESID',float,(nwin,7))
               ]


    @staticmethod
    def _find_tied_index(index, emission_lines):
        i = numpy.where(emission_lines['index'] == index)[0]
        if len(i) > 1:
            raise ValueError('Line indices are not unique!')
        if emission_lines['mode'][i[0]] == 'f':
            return i[0], index
        tied = int(emission_lines['mode'][i[0]][1:])
        j = numpy.where(emission_lines['index'] == tied)[0]
        if len(j) > 1:
            raise ValueError('Line indices are not unique!')
        return (j[0], tied) if emission_lines['mode'][j[0]] == 'f' \
                    else Elric._find_tied_index(tied, emission_lines)

    
    @staticmethod
    def _set_profile_ties(base_profiles, base_restwave, base_fixed_par, tied_profiles,
                          tied_restwave, tied_fixed_par, mode, flux):
        nprof = len(base_profiles)
        if len(base_restwave) != nprof or len(tied_profiles) != nprof \
                or len(tied_restwave) != nprof or len(mode) != nprof or len(flux) != nprof:
            raise ValueError('Must provide same length arrays to _tied_funcs.')
        
        tied_pairs = []
        tied_functions = []
        for i in range(nprof):
            if mode[i][0] == 'w':
                continue 
            tied_pars += [ TiedLineProfile(tied_profiles[i], base_profiles[i]) ]
            tied_fixed_par[i,:] |= base_fixed_par[i,:]
            if mode[i][0] in [ 'a', 'x' ]:
                tied_pairs[i].relative_flux = flux[i]
                tied_functions += [ tied_pairs[i].tie_flux ]
                tied_fixed_par[i,:] |= tied_pairs[i].profile.fix_flux()
            if mode[i][0] in [ 'a', 'k', 'v' ]:
                tied_pairs[i].mean_separation \
                    = (tied_restwave[i]/base_restwave[i]-1)*astropy.constants.c.to('km/s').value
                tied_functions += [ tied_pairs[i].tie_mean ]
                tied_fixed_par[i,:] |= tied_pairs[i].profile.fix_mean()
            if mode[i][0] in [ 'a', 'x', 'k', 's' ]:
                tied_pairs[i].relative_stddev = 1.0
                tied_functions += [ tied_pairs[i].tie_stddev ]
                tied_fixed_par[i,:] |= tied_pairs[i].profile.fix_stddev()

        return numpy.array(tied_pairs), numpy.array(tied_functions), tied_fixed_par


    @staticmethod
    def _line_velocity_offset(restwave, restwave_primary):
        return (restwave/restwave_primary - 1.0) * astropy.constants.c.to('km/s').value


    def _parse_emission_line_models(self):
        r"""

        Parse the input emission-line file into a set of windows --- the
        number of windows is :math:`N_{\rm win}`) --- that are fit for
        each spectrum.
        """

        # Setup the number of windows to fit based on the primary lines
        primary_line = self.emission_lines['mode'] == 'f'
        primary_index = self.emission_lines['index'][primary_line]
        self.nwindows = numpy.sum(primary_line)
        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO,
                       'Number of emission lines to fit: {0}'.format(self.neml))
            log_output(self.loggers, 1, logging.INFO,
                       'Number of fitting windows: {0}'.format(self.nwindows))

        # Set up the primary line data using the guess parameters from
        # the database
        self.fitting_window = numpy.empty(self.nwindows, dtype=object)
        _db_indx = numpy.arange(self.neml)[primary_line]
        for i in range(self.nwindows):

            db_indx = numpy.array([_db_indx[i]])
            line_index = numpy.array([primary_index[i]])
            restwave = numpy.array([ self.emission_lines['restwave'][primary_line][i] ])
            profile_set = numpy.array([ lineprofiles.NCompLineProfile(
                                        self.emission_lines['ncomp'][primary_line][i],
                                        par=self.emission_lines['par'][primary_line][i],
                                        profile=eval('lineprofiles.'
                                            +self.emission_lines['profile'][primary_line][i])) ])
            fixed_par = self.emission_lines['fix'][primary_line][i].astype(bool).reshape(1,-1)
            bounds = numpy.array([ [l,u] \
                                    for l,u in zip(self.emission_lines['lobnd'][primary_line][i],
                                                   self.emission_lines['hibnd'][primary_line][i])
                                 ]).reshape(1,-1,2)
            log_bounds = self.emission_lines['log_bnd'][primary_line][i].astype(bool).reshape(1,-1)
            output_model = self.emission_lines['output_model'][primary_line][i]
            self.fitting_window[i] = ElricFittingWindow(nlines=1, db_indx=db_indx,
                                                        line_index=line_index, restwave=restwave,
                                                        profile_set=profile_set,
                                                        fixed_par=fixed_par, bounds=bounds,
                                                        log_bounds=log_bounds,
                                                        output_model=output_model)

        # Return if nothing is tied
        if self.nwindows == self.neml:
            return

        # Append additional profiles
        for i, e in enumerate(self.emission_lines):
            if primary_line[i]:
                continue
            # Basic check
            if e['index'] == int(e['mode'][1:]):
                raise ValueError('Cannot tie line to itself!')

            # Run up the depenency chain to find the primary line
            # associated with this one
            indx, tied_index = self._find_tied_index(e['index'], self.emission_lines)
            base_indx = numpy.where( primary_index == tied_index)[0][0]

            # Append the line to this fitting window
            self.fitting_window[base_indx].append(i, e['index'], e['restwave'],
                                                  lineprofiles.NCompLineProfile(e['ncomp'],
                                                                                par=e['par'],
                                                                   profile=eval('lineprofiles.'
                                                                                +e['profile'])),
                                                  e['fix'],
                    numpy.array([ [l,u] for l,u in zip(e['lobnd'], e['hibnd']) ]).reshape(-1,2),
                                                  e['log_bnd'], bool(e['output_model']))
            # Set the guess velocity of the profile based on the
            # velocity relative to the main line
            self.fitting_window[base_indx].reset_init_mean(-1,
                            self._line_velocity_offset(self.fitting_window[base_indx].restwave[-1],
                                                       self.fitting_window[base_indx].restwave[0]))

        # Finalize by setting up any tied parameters
        for i in range(self.nwindows):
            if self.fitting_window[i].nlines == 1:
                continue
            # Basic debug check
            if self.emission_lines['mode'][self.fitting_window[i].db_indx[0]] != 'f':
                raise ValueError('DEBUG')

            # The order is important.  Start with the primary line:
            tied = numpy.zeros(self.fitting_window[i].nlines, dtype=bool)
            tied[0] = True              # Primary line already "tied" to the set
            # Iterate until all lines are tied
            while numpy.sum(tied) != self.fitting_window[i].nlines:
                # Find the indices of lines that are directly tied to
                # any previously tied lines.  At the beginning, this
                # only selects lines that are tied to the primary line.
                indx = numpy.where([
                            not tied[j] and numpy.any(int(mode[1:]) \
                                            == self.fitting_window[i].line_index[tied]) \
                            for j,mode in enumerate(
                                self.emission_lines['mode'][self.fitting_window[i].db_indx[:]])
                                   ])[0]

                # For these lines, find the indices in the list of lines
                # for this window with their associated lines to tie to
                tied_indx = [ int(mode[1:]) for mode in \
                                self.emission_lines['mode'][self.fitting_window[i].db_indx[indx]]]
                profile_indx = [ numpy.where(t == self.fitting_window[i].line_index[:])[0][0] \
                                    for t in tied_indx ]

                # Run the maing function that sets up the seried of
                # tying classes/functions and updates the parameters
                # that should be "fixed" during the fit.
                _tied_pairs, _tied_funcs, _fixed_par \
                        = self._set_profile_ties(self.fitting_window[i].profile_set[profile_indx],
                                                 self.fitting_window[i].restwave[profile_indx],
                                numpy.take(self.fitting_window[i].fixed_par, profile_indx, axis=0),
                                                 self.fitting_window[i].profile_set[indx],
                                                 self.fitting_window[i].restwave[indx],
                                numpy.take(self.fitting_window[i].fixed_par, indx, axis=0),
                                self.emission_lines['mode'][self.fitting_window[i].db_indx[indx]],
                                self.emission_lines['flux'][self.fitting_window[i].db_indx[indx]])

                # Add these to the main output
                self.fitting_window[i].tied_pairs = numpy.append(self.fitting_window[i].tied_pairs,
                                                                 _tied_pairs)
                self.fitting_window[i].tied_funcs = numpy.append(self.fitting_window[i].tied_funcs,
                                                                 _tied_funcs)
                for j, jj in enumerate(indx):
                    self.fitting_window[i].fixed_par[jj,:] = _fixed_par[j,:]

                # These lines are now tied!
                tied[indx] = True


    @staticmethod
    def _velocity_vectors(wave, fitting_window):
        return numpy.array([ (wave/fw.restwave[0]-1.0)*astropy.constants.c.to('km/s').value \
                            for fw in fitting_window ])


    @staticmethod
    def _fit_masks(wave, fitting_window, redshift, window_buffer):

        # Input redshift must be a single value
        if not isinstance(redshift, float):
            raise TypeError('redshift must be a single float')

        nwindows = len(fitting_window)
        fitting_mask = numpy.array([ numpy.logical_and(
                                        wave>(fw.restwave[0]-window_buffer)*(1+redshift),
                                        wave<(fw.restwave[0]+window_buffer)*(1+redshift)) \
                                     for fw in fitting_window ])
#        print(fitting_mask.shape)

        for i in range(nwindows):
            if len(fitting_window[i].restwave) == 1:
                continue
            fitting_mask[i,:] |= numpy.any( numpy.array([ numpy.logical_and(
                                                        wave>(rw-window_buffer)*(1+redshift),
                                                        wave<(rw+window_buffer)*(1+redshift)) \
                                                for rw in fitting_window[i].restwave ]), axis=0)
        return fitting_mask


#    @staticmethod
#    def _check_db(emission_lines):
#
#        # Check the input type
#        if not isinstance(emission_lines, EmissionLineDB):
#            raise TypeError('Input database must have type EmissionLineDB.')
#
#        # Check the database itself
#        neml = emission_lines.nsets
#        for i in range(neml):
#            profile = eval(emission_lines['profile'][i])
#            npar = len(profile.param_names)
#            if emission_lines['par'][i].size != npar*emission_lines['ncomp'][i]:
#                raise ValueError('Provided {0} parameters, but expected {1}.'.format(
#                                  emission_lines['par'][i].size, npar*emission_lines['ncomp'][i]))
#            if emission_lines['fix'][i].size != npar*emission_lines['ncomp'][i]:
#                raise ValueError('Provided {0} fix flags, but expected {1}.'.format(
#                                  emission_lines['fix'][i].size, npar*emission_lines['ncomp'][i]))
#            if numpy.any([f not in [0, 1] for f in emission_lines['fix'][i] ]):
#                warnings.warn('Fix values should only be 0 or 1; non-zero values interpreted as 1.')
#            if emission_lines['lobnd'][i].size != npar*emission_lines['ncomp'][i]:
#                raise ValueError('Provided {0} lower bounds, but expected {1}.'.format(
#                                  emission_lines['lobnd'][i].size,
#                                  npar*emission_lines['ncomp'][i]))
#            if emission_lines['hibnd'][i].size != npar*emission_lines['ncomp'][i]:
#                raise ValueError('Provided {0} upper bounds, but expected {1}.'.format(
#                                  emission_lines['hibnd'][i].size,
#                                  npar*emission_lines['ncomp'][i]))
#            if emission_lines['log_bnd'][i].size != npar*emission_lines['ncomp'][i]:
#                raise ValueError('Provided {0} log boundaries designations, but expected '
#                                 '{1}.'.format(emission_lines['log_bnd'][i].size,
#                                               npar*emission_lines['ncomp'][i]))


    @staticmethod
    def _is_near_bound(par, lbnd, ubnd, logbounded, rtol=1e-2, atol=1e-4):
        """
        Determine of any of the parameters within start and start+npar
        are "near" an imposed boundary.
        """
#        print('par: ', par)
#        print('lbnd: ', lbnd)
        lbounded = numpy.invert(numpy.isinf(lbnd))
#        print('L bounded: ', lbounded)

#        print('ubnd: ', ubnd)
        ubounded = numpy.invert(numpy.isinf(ubnd))
#        print('U bounded: ', ubounded)

        if numpy.sum(lbounded) == 0 and numpy.sum(ubounded) == 0:
#            print('No bounds')
            return False

        if numpy.any(lbounded & numpy.invert(ubounded) & (par - lbnd < atol)):
#            print('Near lower bound')
            return True

        if numpy.any(numpy.invert(lbounded) & ubounded & (ubnd - par < atol)):
#            print('Near upper bound')
            return True

        ulbounded = lbounded & ubounded
#        print('UL bounded', ulbounded)
        if numpy.sum(ulbounded) == 0:
#            print('None bounded from both sides')
            return False

#        print('log bounded', logbounded)
#        print('UL & log bounded', ulbounded & logbounded)
#        print('UL & not log bounded', ulbounded & ~logbounded)
        Dp = numpy.zeros(ulbounded.size, dtype=float)
        indx = ulbounded & logbounded
        Dp[indx] = numpy.log10(ubnd[indx]) - numpy.log10(lbnd[indx])
        indx = ulbounded & numpy.invert(logbounded)
        Dp[indx] = ubnd[indx] - lbnd[indx]
        tol = Dp*rtol
#        print(tol)
#        print(tol[ulbounded & logbounded])
#        print((numpy.ma.log10(par) - numpy.ma.log10(lbnd))[ulbounded & logbounded])
#        print((numpy.ma.log10(ubnd) - numpy.ma.log10(par))[ulbounded & logbounded])

        if numpy.any( ulbounded & logbounded & ( (numpy.ma.log10(par) - numpy.ma.log10(lbnd) < tol) 
                                        | (numpy.ma.log10(ubnd) - numpy.ma.log10(par) < tol))):
#            print('Near boundary in log')
            return True
                            
#        print(tol[ulbounded & ~logbounded])
#        print((par - lbnd)[ulbounded & ~logbounded])
#        print((ubnd - par)[ulbounded & ~logbounded])

        if numpy.any(ulbounded & numpy.invert(logbounded) & ((par-lbnd < tol) | (ubnd-par < tol))):
#            print('Near boundary in linear')
            return True
#        print('Not near boundary')
        return False


#        # Check if the parameters are near a boundary
#        if self.bestfit[i,j].bounded:
#            model_fit_par['LOBND'][i,j,:npar] = self.bestfit[i,j].bounds[0]
#            model_fit_par['UPBND'][i,j,:npar] = self.bestfit[i,j].bounds[1]
#            for k in range(npar):
#                if not numpy.isinf(self.bestfit[i,j].bounds[0][k]) \
#                        and model_fit_par['PAR'][i,j,k] - model_fit_par['ERR'][i,j,k] \
#                                    <= self.bestfit[i,j].bounds[0][k]:
#                    near_bound = True
#                    break
#                if not numpy.isinf(self.bestfit[i,j].bounds[1][k]) \
#                        and model_fit_par['PAR'][i,j,k] + model_fit_par['ERR'][i,j,k] \
#                                    >= self.bestfit[i,j].bounds[1][k]:
#                    near_bound = True
#                    break

#            if self.bestfit[i,j].bounded:
#                if numpy.any(~numpy.isinf(self.bestfit[i,j].bounds[0][pl:pl+nlinepar[k]]) \
#                                & (model_fit_par['PAR'][i,j,pl:pl+nlinepar[k]]
#                                    - model_fit_par['ERR'][i,j,pl:pl+nlinepar[k]]
#                                    <= self.bestfit[i,j].bounds[0][pl:pl+nlinepar[k]])):
#                    model_eml_par['MASK'][i,emlj] \
#                            = self.bitmask.turn_on(model_eml_par['MASK'][i,emlj], 'NEAR_BOUND')
#                    near_bound = True
#                if numpy.any(~numpy.isinf(self.bestfit[i,j].bounds[1][pl:pl+nlinepar[k]]) \
#                                & (model_fit_par['PAR'][i,j,pl:pl+nlinepar[k]]
#                                    + model_fit_par['ERR'][i,j,pl:pl+nlinepar[k]] \
#                                    >= self.bestfit[i,j].bounds[1][pl:pl+nlinepar[k]])):
#                    model_eml_par['MASK'][i,emlj] \
#                            = self.bitmask.turn_on(model_eml_par['MASK'][i,emlj], 'NEAR_BOUND')
#                    near_bound = True


    @staticmethod
    def _correct_subeml_par(restwave_0, restwave_k, par, err=None):
        """
        Correct the parameters fit assuming a rest wavelength of
        restwave_0, when it's actually restwave_k.  Input parameters are
        expected to be flux, velocity, velocity dispersion.
        """
        c = astropy.constants.c.to('km/s').value
        _par = par.copy()
        _par[0] *= restwave_k/restwave_0
        _par[1] = (restwave_0*(par[1]/c + 1)/restwave_k - 1)*c
        _par[2] *= restwave_0/restwave_k
        if err is None:
            return _par

        _err = err.copy()
        _err[0] *= restwave_k/restwave_0
        _err[1] *= restwave_0/restwave_k
        _err[2] *= restwave_0/restwave_k
        return _par, _err
        

    def _assess_and_save_fit(self, i, j, model_fit_par, model_eml_par):
        """
        Assess the result of the LineProfileFit results.

            - (DONE) Check the success failure
            - (DONE) Calculate the chi2, rchi2, rms, fractional_rms,
              residuals, fractional_residuals
            - (DONE) Compare the fit parameters to the bounds
            - Check the chi-square and residuals?
            - Check the velocity offset wrt the input
            - For multiple lines, check the order of the lines matches
              the guess parameters
            - For multiple component lines, check the ordering of the
              subcomponents

        """
        # Instantiate the returned flag that the parameter are near the
        # imposed boundary; only meaningful if the fit did not fail
        near_bound = False

        # Assign the fitting window
        model_eml_par['FIT_INDEX'][i,self.fitting_window[j].db_indx] = j
#        print(model_eml_par['FIT_INDEX'][i,:])

        # If the parameters are bounded, save the bounds
        npar = self.bestfit[i,j].npar
        if self.bestfit[i,j].bounded:
            model_fit_par['LOBND'][i,j,:npar] = self.bestfit[i,j].bounds[0]
            model_fit_par['UPBND'][i,j,:npar] = self.bestfit[i,j].bounds[1]

        # Save which parameters were fixed or should be ignored
        model_fit_par['IGNORE'][i,j,npar:] = True
        
#        include_model = True
        if not self.bestfit[i,j].result.success:
            model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                              'FIT_FAILED')
            if not self.quiet:
                log_output(self.loggers, 1, logging.ERROR,
                           'FIT FAILED: Spec: {0}/{1} ({2} : {3})'.format(i+1,self.nspec,
                                                            self.fitting_window[j].line_index,
                                    self.emission_lines['name'][self.fitting_window[j].db_indx]))
            
            return near_bound

        # Get the full list of parameters (including fixed and tied values)
        self.bestfit[i,j]._assign_par(self.bestfit[i,j].result.x)
        par = self.bestfit[i,j].par
        model_fit_par['PAR'][i,j,:npar] = par[:]
        model_fit_par['FIXED'][i,j,:npar] = self.bestfit[i,j].fixed
        model_fit_par['FIXED'][i,j,npar:] = True
        model_fit_par['ERR'][i,j,:npar] = numpy.zeros(npar, dtype=float)
        if self.error is not None:
            tmperr = numpy.diag(self.bestfit[i,j].cov)
            if not numpy.any(tmperr < 0):
                model_fit_par['ERR'][i,j,numpy.invert(model_fit_par['FIXED'][i,j])] \
                        = numpy.sqrt(numpy.diag(self.bestfit[i,j].cov))
            else:
                model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                                  'UNDEFINED_COVAR')


#        print('PAR: ', model_fit_par['PAR'][i,j,:npar])
#        print('ERR: ', model_fit_par['ERR'][i,j,:npar])

        # Get the per-line parameters
        nlinepar = numpy.array([ p.npar for p in self.fitting_window[j].profile_set ])
        for k in range(self.fitting_window[j].nlines):
            pl = 0 if k == 0 else numpy.sum(nlinepar[:k])
            emlj = self.fitting_window[j].db_indx[k]
            model_eml_par['FLUX'][i,emlj] \
                    = self.fitting_window[j].profile_set[k].moment(order=0,
                                                                   par=par[pl:pl+nlinepar[k]])

            model_eml_par['FLUXERR'][i,emlj] \
                    = self.fitting_window[j].profile_set[k].moment_err(order=0,
                                                                       par=par[pl:pl+nlinepar[k]],
                                                    err=model_fit_par['ERR'][i,j,pl:pl+nlinepar[k]])
            # Flux unit conversion
#            print(astropy.constants.c.to('km/s').value/self.fitting_window[j].restwave[k])
#                    /= astropy.constants.c.to('km/s').value/self.fitting_window[j].restwave[k]
            model_eml_par['FLUX'][i,emlj] \
                    *= self.fitting_window[j].restwave[0]/astropy.constants.c.to('km/s').value
            model_eml_par['FLUXERR'][i,emlj] \
                    *= self.fitting_window[j].restwave[0]/astropy.constants.c.to('km/s').value

            model_eml_par['KIN'][i,emlj,0] \
                    = self.fitting_window[j].profile_set[k].moment(order=1,
                                                                   par=par[pl:pl+nlinepar[k]])
            model_eml_par['KINERR'][i,emlj,0] \
                    = self.fitting_window[j].profile_set[k].moment_err(order=1,
                                                                       par=par[pl:pl+nlinepar[k]],
                                                    err=model_fit_par['ERR'][i,j,pl:pl+nlinepar[k]])

            model_eml_par['KIN'][i,emlj,1] \
                    = self.fitting_window[j].profile_set[k].moment(order=2,
                                                                   par=par[pl:pl+nlinepar[k]])

            model_eml_par['KINERR'][i,emlj,1] \
                    = self.fitting_window[j].profile_set[k].moment_err(order=2,
                                                                       par=par[pl:pl+nlinepar[k]],
                                                    err=model_fit_par['ERR'][i,j,pl:pl+nlinepar[k]])

            # Correct for the fact that the fit was done using a
            # velocity vector defined by the primary line
            # TODO: Need to change how fitting is done so that this is
            # NOT necessary!
            if k > 0:
                line_par, line_parerr \
                        = self._correct_subeml_par(self.fitting_window[j].restwave[0],
                                                   self.fitting_window[j].restwave[k],
                                                   numpy.array([ model_eml_par['FLUX'][i,emlj],
                                                                 model_eml_par['KIN'][i,emlj,0],
                                                                 model_eml_par['KIN'][i,emlj,1] ]),
                                                   err= numpy.array([
                                                            model_eml_par['FLUXERR'][i,emlj],
                                                            model_eml_par['KINERR'][i,emlj,0],
                                                            model_eml_par['KINERR'][i,emlj,1] ]))
                model_eml_par['FLUX'][i,emlj], model_eml_par['KIN'][i,emlj,0], \
                    model_eml_par['KIN'][i,emlj,1] = tuple(line_par)
                model_eml_par['FLUXERR'][i,emlj], model_eml_par['KINERR'][i,emlj,0], \
                    model_eml_par['KINERR'][i,emlj,1] = tuple(line_parerr)

            if self.bestfit[i,j].bounded \
                    & self._is_near_bound(model_fit_par['PAR'][i,j,pl:pl+nlinepar[k]],
                                          self.bestfit[i,j].bounds[0][pl:pl+nlinepar[k]],
                                          self.bestfit[i,j].bounds[1][pl:pl+nlinepar[k]],
                            self.fitting_window[j].log_bounds.ravel().copy()[pl:pl+nlinepar[k]]):
                model_eml_par['MASK'][i,emlj] \
                            = self.bitmask.turn_on(model_eml_par['MASK'][i,emlj], 'NEAR_BOUND')
                near_bound = True

#        print('Any line in window near boundary: ', near_bound)
        if near_bound:
            model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                              'NEAR_BOUND')

        # Set the fit statistics
        model_fit_par['CHI2'][i,j] = 0.0 if self.error is None else \
                                            numpy.sum(numpy.square(self.bestfit[i,j]._chi(par)))
#        print('CHI2:', model_fit_par['CHI2'][i,j])
        model_fit_par['RCHI2'][i,j] = model_fit_par['CHI2'][i,j] \
                                            / (model_fit_par['NPIXFIT'][i,j] - npar)
#        print('RCHI2:', model_fit_par['RCHI2'][i,j])
        resid = self.bestfit[i,j]._resid(par)
        model_fit_par['RMS'][i,j] = numpy.sqrt(numpy.mean(numpy.square(resid)))
#        print('RMS:', model_fit_par['RMS'][i,j])

        model_fit_par['RESID'][i,j] = sample_growth(numpy.ma.absolute(resid),
                                                    [0.0, 0.25, 0.50, 0.75, 0.90, 0.99, 1.0])
#        model_fit_par['RESID'][i,j] = residual_growth(resid, [0.25, 0.50, 0.75, 0.90, 0.99])
#        print('RESID: ', model_fit_par['RESID'][i,j])

        fit = self.bestfit[i,j].sample(self.bestfit[i,j].x, par)
        indx = numpy.absolute(fit) > 1e-4
        if numpy.sum(indx) > 0:
            frac_resid = resid[indx]/fit[indx]
            model_fit_par['FRAC_RMS'][i,j] = numpy.sqrt(numpy.mean(numpy.square(frac_resid)))
#            print('FRMS:', model_fit_par['FRAC_RMS'][i,j])
            if numpy.sum(indx) > 1:
                model_fit_par['FRAC_RESID'][i,j] = sample_growth(numpy.ma.absolute(frac_resid),
                                                        [0.0, 0.25, 0.50, 0.75, 0.90, 0.99, 1.0])
#                model_fit_par['FRAC_RESID'][i,j] = residual_growth(frac_resid,
#                                                               [0.25, 0.50, 0.75, 0.90, 0.99])
#                print('FRAC_RESID: ', model_fit_par['FRAC_RESID'][i,j])

        return near_bound


#    def _instrumental_dispersion_correction(self, model_eml_par):
#        """
#        Determine the instrumental velocity dispersion at the centroids
#        of the fitted emission lines.
#        """
#        if self.sres is None:
#            return
#
#        interpolator = interpolate.interp1d(self.wave, self.sres, fill_value='extrapolate',
#                                            assume_sorted=True)
#        restwave = numpy.array([self.emission_lines['restwave']]*self.nspec)
#        cnst = constants()
#
#        # Get the instrumental dispersion at each (valid) line center
#        indx = ~self.bitmask.flagged(model_eml_par['MASK'],
#                                     flag=['INSUFFICIENT_DATA', 'FIT_FAILED', 'NEAR_BOUND',
#                                           'UNDEFINED_COVAR' ])
#        model_eml_par['SINST'][indx] = astropy.constants.c.to('km/s') \
#                                            / interpolator((model_eml_par['KIN'][indx,0] 
#                                                        / astropy.constants.c.to('km/s').value+1.0)
#                                                                *restwave[indx])/cnst.sig2fwhm

#        z = model_eml_par['KIN'][indx,0] / astropy.constants.c.to('km/s').value
#        pyplot.scatter(restwave[indx]*(1.0+z), model_eml_par['SINST'][indx], marker='.', s=30,
#                       color='k')
#        pyplot.show()
#
#        pyplot.scatter(model_eml_par['SINST'][indx], model_eml_par['KIN'][indx,1], marker='.',
#                       s=30, color='k')
#        pyplot.show()


# NEVER TESTED
#    def _correct_velocity_dispersion(self, model_eml_par):
#        """
#        Use the **previously calculated** instrumental velocity
#        dispersion to correct the measured velocity dispersions of the
#        lines to the astrophysical velocity dispersion.
#        """
#
#        # Correct the second moments for the instrumental dispersion
#        defined = numpy.zeros(indx.shape, dtype=numpy.bool)
#        defined[indx] = model_eml_par['SINST'][indx] < model_eml_par['KIN'][indx,1]
#        nonzero = numpy.zeros(indx.shape, dtype=numpy.bool)
#        nonzero[indx] = numpy.absolute(model_eml_par['SINST'][indx] 
#                                            - model_eml_par['KIN'][indx,1]) > 0
#
##        orig = model_eml_par['KIN'].copy()
#
#        model_eml_par['KINERR'][nonzero,1] \
#                    = 2.0*model_eml_par['KIN'][nonzero,1]*model_eml_par['KINERR'][nonzero,1]
#        model_eml_par['KIN'][nonzero,1] = numpy.square(model_eml_par['KIN'][nonzero,1]) \
#                                                - numpy.square(model_eml_par['SINST'][nonzero])
#        model_eml_par['KIN'][nonzero,1] = model_eml_par['KIN'][nonzero,1] \
#                            / numpy.sqrt(numpy.absolute(model_eml_par['KIN'][nonzero,1]))
#        model_eml_par['KINERR'][nonzero,1] /= numpy.absolute(2.*model_eml_par['KIN'][nonzero,1])
#
#        # Flag undefined values
#        model_eml_par['MASK'][indx & ~defined] \
#                = self.bitmask.turn_on(model_eml_par['MASK'][indx & ~defined], 'UNDEFINED_SIGMA')
#
##        flg = self.bitmask.flagged(model_eml_par['MASK'], flag='UNDEFINED_SIGMA')
##        pyplot.scatter(orig[nonzero & flg,1], model_eml_par['KIN'][nonzero & flg,1], marker='.',
##                       s=30, color='0.8')
##        pyplot.scatter(orig[nonzero & ~flg,1], model_eml_par['KIN'][nonzero & ~flg,1], marker='.',
##                       s=30, color='k')
##        pyplot.show()

    def fit_SpatiallyBinnedSpectra(self, binned_spectra, par=None, loggers=None, quiet=False):
        """

        This is a basic interface that is geared for the DAP that
        interacts with the rest of the, more general, parts of the
        class.

        This should not declare anything to self!

        """
        # Assign the parameters if provided
        if par is None:
            raise ValueError('Required parameters for LineProfileFit have not been defined.')

        # SpatiallyBinnedSpectra object always needed
        if binned_spectra is None:
            raise ValueError('Must provide spectra object for fitting.')
        if not isinstance(binned_spectra, SpatiallyBinnedSpectra):
            raise TypeError('Must provide a valid SpatiallyBinnedSpectra object!')
        if binned_spectra.hdu is None:
            raise ValueError('Provided SpatiallyBinnedSpectra object is undefined!')

        # Continuum accounts for underlying absorption
        if par['stellar_continuum'] is not None \
                and not isinstance(par['stellar_continuum'], StellarContinuumModel):
            raise TypeError('Must provide a valid StellarContinuumModel object.')
        continuum = None if par['stellar_continuum'] is None \
                        else par['stellar_continuum'].fill_to_match(binned_spectra['BINID'].data,
                                                            missing=binned_spectra.missing_bins)
        if continuum is not None:
            if continuum.shape != binned_spectra['FLUX'].data.shape:
                raise ValueError('Provided continuum does not match shape of the binned spectra.')
            if not isinstance(continuum, numpy.ma.MaskedArray):
                continuum = numpy.ma.MaskedArray(continuum)

        # Get the data arrays to fit
        good_snr = binned_spectra.above_snr_limit(par['minimum_snr'])
        wave, flux, ivar, sres = EmissionLineFit.get_spectra_to_fit(binned_spectra,
                                                                    pixelmask=par['pixelmask'],
                                                                    select=good_snr)

        # Return the fitted data
        model_wave, model_flux, model_base, model_mask, model_fit_par, model_eml_par \
                = self.fit(binned_spectra['WAVE'].data, flux, par['emission_lines'],
                           ivar=ivar, sres=sres, continuum=continuum[good_snr,:],
                           base_order=par['base_order'], window_buffer=par['window_buffer'],
                           guess_redshift=par['guess_redshift'][good_snr],
                           guess_dispersion=par['guess_dispersion'][good_snr], loggers=loggers,
                           quiet=quiet)

        # Save the the bin ID numbers indices based on the spectra
        # selected to be fit
        model_fit_par['BINID'] = binned_spectra['BINS'].data['BINID'][good_snr]
        model_fit_par['BINID_INDEX'] = numpy.arange(binned_spectra.nbins)[good_snr]

        model_eml_par['BINID'] = binned_spectra['BINS'].data['BINID'][good_snr]
        model_eml_par['BINID_INDEX'] = numpy.arange(binned_spectra.nbins)[good_snr]

        # Add the equivalent width data
        EmissionLineFit.measure_equivalent_width(binned_spectra['WAVE'].data, flux,
                                                 par['emission_lines'], model_eml_par,
                                                 redshift=par['guess_redshift'][good_snr],
                                                 bitmask=self.bitmask)

        # Only return model and model parameters for the *fitted*
        # spectra
        return model_flux, model_base, model_mask, model_fit_par, model_eml_par, None


    def fit(self, wave, flux, emission_lines, ivar=None, mask=None, sres=None,
            continuum=None, base_order=-1, window_buffer=25, guess_redshift=None,
            guess_dispersion=None, loggers=None, quiet=False):
        """
        The flux array is expected to have size Nspec x Nwave.

        Raises:
            ValueError: Raised if the length of the spectra, errors, or
                mask does not match the length of the wavelength array;
                raised if the wavelength, redshift, or dispersion arrays
                are not 1D vectors; and raised if the number of
                redshifts or dispersions is not a single value or the
                same as the number of input spectra.
        """
        # Initialize the reporting
        if loggers is not None:
            self.loggers = loggers
        self.quiet = quiet

        # Check the input emission-line database
        EmissionLineFit.check_emission_line_database(emission_lines)
        self.emission_lines = emission_lines
        self.neml = self.emission_lines.neml

        # Prepare the spectra for fitting
        self.wave = wave
        self.flux, self.error, self.sres, self.continuum, self.redshift, self.dispersion \
                    = EmissionLineFit.check_and_prep_input(self.wave, flux, ivar=ivar, mask=mask,
                                                           sres=sres, continuum=continuum,
                                                           redshift=guess_redshift,
                                                           dispersion=guess_dispersion,
                                                        default_dispersion=self.default_dispersion)
        # Keep the shape
        self.nspec, self.nwave = self.flux.shape
        # Subtract the continuum
        self.fluxnc, self.no_continuum = EmissionLineFit.subtract_continuum(self.flux,
                                                                            self.continuum)
        # No continuum for any spaxel!
        if self.no_continuum is None:
            self.no_continuum = numpy.ones(self.flux.shape, dtype=bool)

#        pyplot.step(self.wave, self.flux[0,:], where='mid', color='k', lw=0.5, linestyle='-')
#        pyplot.step(self.wave, self.fluxnc[0,:], where='mid', color='g', lw=0.5, linestyle='-')
#        pyplot.plot(self.wave, self.no_continuum[0,:], color='r', lw=0.5, linestyle='-')
#        pyplot.show()
#        exit()

        # Build the emission-line fitting windows
        self._parse_emission_line_models()
        max_npar = numpy.amax([ fw.fixed_par.size for fw in self.fitting_window ])
        self.base_order = base_order
        if self.base_order > -1:
            max_npar += base_order+1
        self.window_buffer = window_buffer

        # Initialize the output arrays
        #  Model flux:
        model_flux = numpy.zeros(self.flux.shape, dtype=float)
        model_base = numpy.zeros(self.flux.shape, dtype=float)
        #  Model mask:
        model_mask = numpy.zeros(self.flux.shape, dtype=self.bitmask.minimum_dtype())
        indx = numpy.ma.getmaskarray(self.flux)
        model_mask[indx] = self.bitmask.turn_on(model_mask[indx], 'DIDNOTUSE')
        model_mask[self.no_continuum] = self.bitmask.turn_on(model_mask[self.no_continuum],
                                                             'NOCONTINUUM')

        # Model parameters and fit quality
        model_fit_par = init_record_array(self.nspec,
                                          Elric._per_fitting_window_dtype(self.nwindows, max_npar,
                                                                self.bitmask.minimum_dtype()))
        model_fit_par['BINID'] = numpy.arange(self.nspec)
        model_fit_par['BINID_INDEX'] = numpy.arange(self.nspec)

        # Emission-line parameters
        model_eml_par = self.emission_line_datatable(self.neml, 2, self.bitmask.minimum_dtype(),
                                                     shape=self.nspec)
#        model_eml_par = init_record_array(self.nspec,
#                                          self._per_emission_line_dtype(self.neml, 2,
#                                                                self.bitmask.minimum_dtype()))
        model_eml_par['CONTMPLY'] = numpy.ones(model_eml_par['CONTMPLY'].shape, dtype=float)
        model_eml_par['BINID'] = numpy.arange(self.nspec)
        model_eml_par['BINID_INDEX'] = numpy.arange(self.nspec)

        t = time.perf_counter()
        # Do the fit
        velocity = self._velocity_vectors(self.wave, self.fitting_window)
        self.bestfit = numpy.empty((self.nspec,self.nwindows), dtype=object)
        for i in range(self.nspec):
#        for i in range(5):

#            for j in range(self.nwindows):
#                print('Window: {0}/{1}'.format(j+1, self.nwindows))
#                for p in self.fitting_window[j].profile_set:
#                    print(p.par)
#            if not self.quiet:
#                log_output(self.loggers, 1, logging.INFO, 'Fit: {0}/{1}'.format(i+1,self.nspec))
            print('Fit: {0}/{1}'.format(i+1,self.nspec), end='\r')

            # Get the fitting mask for each emission-line in each window
            fitting_mask = self._fit_masks(self.wave, self.fitting_window, self.redshift[i],
                                           self.window_buffer)

            # Set the mask for any pixels that are NEVER fit for this
            # spectrum
            indx = numpy.invert(numpy.any(fitting_mask, axis=0))
            model_mask[i,indx] = self.bitmask.turn_on(model_mask[i,indx], 'OUTSIDE_RANGE')

            model_jump = numpy.array([
                        numpy.sum(fm & self.no_continuum[i,:]) not in [0, numpy.sum(fm)] \
                        for fm in numpy.take(fitting_mask, numpy.arange(self.nwindows), axis=0) ])
            spec_to_fit = numpy.ma.array([ self.flux[i,:] if mj else self.fluxnc[i,:] \
                                            for mj in model_jump ])

            cz = self.redshift[i]*astropy.constants.c.to('km/s').value

#            total_near_bound = 0
            # Fit each window
            for j in range(self.nwindows):
                # Rest the profile to the initial parameters
                self.fitting_window[j].reinit_profiles()
#                print('Fitting: ', self.emission_lines['name'][self.fitting_window[j].db_indx])

                # No observed pixels in fitting window
#                print('Sum of fitting mask: {0}'.format(numpy.sum(fitting_mask[j,:])))
                if numpy.sum(fitting_mask[j,:]) == 0:
                    warnings.warn('No valid data for fit to line(s) {0} ({1}).  '
                                  'Continuuing'.format(self.fitting_window[j].line_index,
                                    self.emission_lines['name'][self.fitting_window[j].db_indx]))
                    model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                                      'INSUFFICIENT_DATA')
                    continue

#                w,h = pyplot.figaspect(1)
#                fig = pyplot.figure(figsize=(1.5*w,1.5*h))
#                ax = fig.add_axes([0.1, 0.3, 0.8, 0.4])
#                ax.set_xlim(velocity[j,0], velocity[j,-1])
#                ax.step(velocity[j,:], spec_to_fit[j,:], where='mid', color='k', linestyle='-',
#                        lw=0.5)
#                ax.plot(velocity[j,:], fitting_mask[j,:], color='r')
#                axt = ax.twiny()
#                axt.set_xlim(self.wave[0], self.wave[-1])
#                pyplot.show()

                # Shift the guess velocity based on the redshift and
                # adjust the guess flux
                # TODO: Use emission-line moment results for the
                # guesses, if provided
                # TODO: Use self.dispersion to set the initial
                # dispersion guesses?  Currently uses guesses from
                # emission-line database (see
                # _parse_emission_line_models)
                for k,p in enumerate(self.fitting_window[j].profile_set):
                    p.shift_mean(cz)
                    vi = numpy.argsort(numpy.absolute(velocity[j,:]-p.moment(order=1)))[0]
#                    print('guess flux: ', (spec_to_fit[j,vi] if spec_to_fit[j,vi] > 0.1 else 0.1) \
#                                * numpy.sqrt(2*numpy.pi) * p.moment(order=2))
                    p.set_flux((spec_to_fit[j,vi] if spec_to_fit[j,vi] > 0.1 else 0.1) \
                                * numpy.sqrt(2*numpy.pi) * p.moment(order=2))

                # Setup the guess parameters and bounds
                _guess_par = numpy.array([ p.par.copy().ravel() \
                                            for p in self.fitting_window[j].profile_set ]).ravel()
                _fixed_par = self.fitting_window[j].fixed_par.ravel().copy()
                _bounds = self.fitting_window[j].bounds.copy().reshape(-1,2)
#                print(_guess_par)
#                print(_fixed_par)
#                print(_bounds)

                # For individual lines, bound the line mean to be within
                # the fitting window
                if self.fitting_window[j].nlines == 1:
                    indx = self.fitting_window[j].profile_set[0].mean_indx()
                    _bounds[indx,0] = velocity[j,fitting_mask[j,:]][0]
                    _bounds[indx,1] = velocity[j,fitting_mask[j,:]][-1]
                # For multiple lines, bound the line mean to be within
                # its own region of the window defined by the +/- 2/3 of
                # the velocity separation to the nearest line.  These
                # borders overlap!
                else:
                    indx = numpy.array(self.fitting_window[j].profile_set[0].mean_indx())
                    for k in range(1,self.fitting_window[j].nlines):
                        indx = numpy.append(indx,
                                numpy.array(self.fitting_window[j].profile_set[0].mean_indx()) \
                            + sum([self.fitting_window[j].profile_set[kk].npar for kk in range(k)]))
                    srt = numpy.argsort(_guess_par[indx])
                    borders = []
                    for k in range(len(indx)-1):
                        diff = _guess_par[indx][srt[k+1]]-_guess_par[indx][srt[k]]
                        borders += [ _guess_par[indx][srt[k]] + 2.0*diff/3.0,
                                     _guess_par[indx][srt[k]] + diff/3.0 ]
                    borders = numpy.array([ velocity[j,fitting_mask[j,:]][0], *borders,
                                            velocity[j,fitting_mask[j,:]][-1] ])
                    _bounds[indx,0] = borders[2*srt]
                    _bounds[indx,1] = borders[2*srt+1]

                # Add the parameters for the baseline
                if base_order > -1:
                    _guess_par = numpy.append(_guess_par, numpy.zeros(base_order+1))
#                    _guess_par[-base_order-1] = numpy.median(spec_to_fit[j,fitting_mask[j,:]])
#                    print(0.0, _guess_par[-base_order-1])
#                    pyplot.scatter(velocity[j,fitting_mask[j,:]], spec_to_fit[j,fitting_mask[j,:]],
#                                   marker='.', s=100, color='k')
#                    y = velocity.copy()[j,fitting_mask[j,:]]
#                    y[:] = numpy.median(spec_to_fit[j,fitting_mask[j,:]])
#                    pyplot.plot(velocity[j,fitting_mask[j,:]], y, color='r')
#                    pyplot.show()
#                    exit()
                    _fixed_par = numpy.append(self.fitting_window[j].fixed_par.ravel(),
                                              numpy.zeros(base_order+1).astype(bool))
                    _bounds = numpy.append(_bounds,
                          numpy.array([-numpy.inf,numpy.inf]*(base_order+1)).reshape(-1,2),axis=0)

                # Check there is sufficient data to perform the fit
                model_fit_par['NPIXFIT'][i,j] = len(spec_to_fit[j,fitting_mask[j,:]].compressed())
                if model_fit_par['NPIXFIT'][i,j] <= _guess_par.size:
                    warnings.warn('Insufficient data points ({0}) to fit with this function ' \
                                  '(npar={1})!'.format(
                                        len(spec_to_fit[j,fitting_mask[j,:]].compressed()),
                                        _guess_par.size))
                    self.bestfit[i,j] = None
                    # Flag that the fit was not performed
                    model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                                      'INSUFFICIENT_DATA')
                    continue

                # Make sure that the guess is still in the bounds:
                # TODO: How do I deal with this:
                #   - throw a warning and adjust the guess
                #   - as above and include a mask bit
                #   - don't perform the fit because it will be bogus
                indx = numpy.logical_or( _guess_par < _bounds[:,0], _guess_par > _bounds[:,1])
                if numpy.sum(indx) > 0:
                    warnings.warn('Initial guess outside bounds for line(s) {0} ({1}).  '
                                  'Adjusted to center of bounds.'.format(
                                    self.fitting_window[j].line_index,
                                    self.emission_lines['name'][self.fitting_window[j].db_indx]))
                    _guess_par[indx] = numpy.mean(_bounds[indx,:], axis=1)

                # TODO: Get rid of this debugging issue...
                if not numpy.all(_bounds[:,0]<_bounds[:,1]):

                    print('input')
                    print(_bounds)
                    print(self.fitting_window[j].bounds.reshape(-1,2))
                    print(numpy.sum(fitting_mask[j,:]))

                    if self.fitting_window[j].nlines == 1:
                        indx = self.fitting_window[j].profile_set[0].mean_indx()
                        _bounds[indx,0] = velocity[j,fitting_mask[j,:]][0]
                        _bounds[indx,1] = velocity[j,fitting_mask[j,:]][-1]
                    else:
                        indx = numpy.array(self.fitting_window[j].profile_set[0].mean_indx())
                        for k in range(1,self.fitting_window[j].nlines):
                            indx = numpy.append(indx,
                                numpy.array(self.fitting_window[j].profile_set[0].mean_indx()) \
                                + sum([self.fitting_window[j].profile_set[kk].npar
                                    for kk in range(k)]))
                        srt = numpy.argsort(_guess_par[indx])
                        borders = []
                        for k in range(len(indx)-1):
                            diff = _guess_par[indx][srt[k+1]]-_guess_par[indx][srt[k]]
                            borders += [ _guess_par[indx][srt[k]] + 2.0*diff/3.0,
                                        _guess_par[indx][srt[k]] + diff/3.0 ]
                        borders = numpy.array([ velocity[j,fitting_mask[j,:]][0], *borders,
                                                velocity[j,fitting_mask[j,:]][-1] ])
                        print('borders')
                        print(borders)
                        _bounds[indx,0] = borders[2*srt]
                        _bounds[indx,1] = borders[2*srt+1]

                    print('output')
                    print(_bounds)
                    raise ValueError('WTF')

#                print(_guess_par)
#                print(_bounds)
#                pyplot.errorbar(velocity[j,fitting_mask[j,:]], spec_to_fit[j,fitting_mask[j,:]],
#                                yerr=_error[i,fitting_mask[j,:]], marker='.', color='k',
#                                linestyle='', capsize=0)
#                pyplot.title('Spec: {0}; Line (set): {1}'.format(i+1,j+1))
#                pyplot.show()
#                print('N valid pixels: {0}'.format(
#                                               len(spec_to_fit[j,fitting_mask[j,:]].compressed())))
                
                # Run the fit
#                print('guess par: {0} {1}'.format(i, j))
#                print(_guess_par)
#                print(' ')
                self.bestfit[i,j] = LineProfileFit(velocity[j,:], spec_to_fit[j,:],
                                                   self.fitting_window[j].profile_set.tolist(),
                                            error=None if self.error is None else self.error[i,:],
                                                   base_order=base_order,
                                                   mask=numpy.invert(fitting_mask[j,:]),
                                                   par=_guess_par,
                                                   bounds=(_bounds[:,0], _bounds[:,1]),
                                                   construct_covariance=True)
                                                   #, verbose=2)#, fixed=_fixed_par)
#                print('Spec: {0}; Line (set): {1}; Success: {2}'.format(i+1,j+1,
#                                                                self.bestfit[i,j].result.success))

                near_bound = self._assess_and_save_fit(i, j, model_fit_par, model_eml_par)
                if near_bound:
#                    print('Flagging mask as near bound')
                    model_mask[i,fitting_mask[j,:]] \
                            = self.bitmask.turn_on(model_mask[i,fitting_mask[j,:]], 'NEAR_BOUND')
#                    total_near_bound += 1

                # Add the best-fit to the lines to the composite model
                # for this spectrum
                if self.fitting_window[j].output_model and not self.bestfit[i,j].result.success:
                    
                    model_mask[i,fitting_mask[j,:]] \
                            = self.bitmask.turn_on(model_mask[i,fitting_mask[j,:]], 'FIT_FAILED')
                elif not self.fitting_window[j].output_model or model_fit_par['MASK'][i,j] != 0:
                    model_fit_par['MASK'][i,j] = self.bitmask.turn_on(model_fit_par['MASK'][i,j],
                                                                      'EXCLUDED_FROM_MODEL')
                elif self.fitting_window[j].output_model and model_fit_par['MASK'][i,j] == 0:
                    # Get the full model (eventually will only be the
                    # baseline)
                    p = self.bestfit[i,j].result.x.copy()
                    model_base[i,fitting_mask[j,:]] \
                            += self.bestfit[i,j].sample(velocity[j,fitting_mask[j,:]], par=p)
                    # Remove the baseline and only get the line profile
                    p[-base_order-1:] = 0.0
                    model_flux[i,fitting_mask[j,:]] \
                            += self.bestfit[i,j].sample(velocity[j,fitting_mask[j,:]], par=p)

#                if 44 in self.fitting_window[j].line_index:
#                    pyplot.errorbar(velocity[j,fitting_mask[j,:]], spec_to_fit[j,fitting_mask[j,:]],
#                                    yerr=self.error[i,fitting_mask[j,:]], marker='.', color='k',
#                                    linestyle='', capsize=0)
#                    pyplot.plot(velocity[j,fitting_mask[j,:]],
#                                self.bestfit[i,j].sample(velocity[j,fitting_mask[j,:]],
#                                par=self.bestfit[i,j].result.x), linestyle='-', color='g')
#                    pyplot.plot(velocity[j,fitting_mask[j,:]],
#                                self.bestfit[i,j].sample(velocity[j,fitting_mask[j,:]],
#                                par=_guess_par), linestyle='-', color='r')
#                    print(_guess_par)
#                    print(self.bestfit[i,j].result.x)
#                    pyplot.title('Spec: {0}; Line (set): {1}'.format(i+1,j+1))
#                    pyplot.show()

#            pyplot.plot(wave, model_flux[i,:], linestyle='-', color='g')
#            pyplot.plot(wave, model_base[i,:], linestyle='-', color='r')
#            pyplot.show()

#            print('Windows with parameters near boundary: {0}'.format(total_near_bound))

            # Determine the instrumental dispersion correction at the line centers
            indx = numpy.invert(self.bitmask.flagged(model_eml_par['MASK'][i,:],
                                                     flag=['INSUFFICIENT_DATA', 'FIT_FAILED',
                                                           'NEAR_BOUND', 'UNDEFINED_COVAR' ]))
            model_eml_par['SIGMACORR'][i,indx] \
                    = EmissionLineFit.instrumental_dispersion(self.wave, self.sres[i,:],
                                                              emission_lines['restwave'][indx],
                                                              model_eml_par['KIN'][i,indx,0])

        # With these definitions, the instrumental sigma and the sigma correction are
        # identical, and the "template" sigma are all zero
        model_eml_par['SIGMAINST'] = model_eml_par['SIGMACORR'].copy()

        print('Fit: {0}/{0}'.format(i+1,self.nspec))
        # Remove the lines from the full model to just provide the
        # baseline model
        model_base -= model_flux

#        pyplot.plot(wave, model_flux[0,:], linestyle='-', color='g')
#        pyplot.plot(wave, model_base[0,:], linestyle='-', color='r')
#        pyplot.show()

        if not self.quiet:
            log_output(self.loggers, 1, logging.INFO, 'Fits completed in {0:.4e} min.'.format(
                       (time.perf_counter() - t)/60))

        return self.wave, model_flux, model_base, model_mask, model_fit_par, model_eml_par





# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-
"""
Provides a set of functions to handle and alter the instrumental
resolution.

----

.. include license and copyright
.. include:: ../include/copy.rst

----

.. include common links, assuming primary doc root is up one directory
.. include:: ../include/links.rst
"""
import warnings

from IPython import embed

import numpy

from scipy import interpolate
from scipy.special import erf

import astropy.constants
from .constants import DAPConstants

from .sampling import spectrum_velocity_scale

#from matplotlib import pyplot

class VariableGaussianKernel:
    r"""
    Support class for variable sigma convolution.  See
    :func:`convolution_variable_sigma`.

    Modest modification of M. Cappellari's
    :func:`ppxf.ppxf_util.gaussian_filter1d` function.

    Args:
        sigma (numpy.ndarray):
            Coordinate-dependent standard deviation of the Gaussian
            kernel in pixels.
        minsig (:obj:`float`, optional):
            The minimum sigma value allowed.
        nsig (:obj:`float`, optional):
            The extent of the calculation of the kernel for the
            convolution.  By default, the kernel is not calculated
            beyond 3*sigma.
        integral (:obj:`bool`, optional):
            Construct the kernel based on the integral of the Gaussian
            over the pixel width using the error function, instead of
            just sampling the Gaussian function directly.

    Raises:
        ValueError:
            Raised if the input sigma has no length.

    Attributes:
        sigma (numpy.ndarray):
            Coordinate-dependent standard deviation of the Gaussian
            kernel.
        p (int):
            The maximum number of pixels on either side of the kernel
            center needed for the convolution.
        m (int):
            Number of pixels in the kernel (:math:`m = 2p + 1`).
        kernel (numpy.ndarray):
            The calculated kernel to use for the convolution.
    """
    def __init__(self, sigma, minsig=0.01, nsig=3.0, integral=False):
        self.n = sigma.size                                     # Vector length
        if self.n == 0:
            raise ValueError('Input sigma has zero length!')
        self.sigma = sigma.clip(min=minsig)                     # Force sigmas to minimum 
        self.p = int(numpy.ceil(numpy.amax(nsig*self.sigma)))   # Kernel covers up to nsig*sigma
        self.m = 2*self.p + 1                                   # Kernel length
        x = numpy.linspace(-self.p, self.p, self.m)             # X^2 for kernel
        x2 = numpy.square(x)

        # Kernel will have size m x n
        self.kernel = (erf((x[:,None]+0.5)/numpy.sqrt(2)/self.sigma) 
                            - erf((x[:,None]-0.5)/numpy.sqrt(2)/self.sigma))/2. if integral else \
                      numpy.exp(-x2[:, None]/(2*numpy.square(self.sigma))) \
                                / numpy.sqrt(2*numpy.pi) / self.sigma

        self.kernel /= numpy.sum(self.kernel, axis=0)[None, :]       # Normalize kernel


    def _check_shape(self, y, ye=None):
        """
        Make sure that the shapes are appropriate for the defined kernel.
        """
        if len(y.shape) != 1:
            raise ValueError('y must be a 1D array!')
        if y.size != self.n:
            raise ValueError('y and sigma must have the same shape!')
        if ye is not None and (len(ye.shape) != 1 or ye.size != self.n):
            raise ValueError('ye length does not must have the correct shape!')


    def _create_a(self, y):
        a = numpy.zeros(self.kernel.shape)
        for i in range(self.m):
            a[i,self.p:-self.p] = y[i:self.n-self.m+i+1]
        return a


    def convolve(self, y, ye=None):
        """
        Convolve the input vector with the variable-width Gaussian.

        The input `y` must be a 1D vector with the same length as
        :attr:`sigma`.  If errors are provided, they must also adhere to
        these shape limitations and are propagated to the convolved
        vector.

        Args:
            y (numpy.ndarray):
                Vector to convolve
            ye (numpy.ndarray, optional):
                Error in the vector to convolve
    
        Raises:
            ValueError:
                Raised if *y* is not a 1D vector, or if the shape of *y*
                and *sigma* (and *ye* if provided) are different.

        Returns:
            `numpy.ndarray`_: The convolved vector. If the errors are
            provided, the function returns two vectors, the convolved
            vector and its error.
        """
        self._check_shape(y, ye=ye)

        # Create m copies of the shifted input function
        a = self._create_a(y)
        if ye is None:
            return numpy.sum(a*self.kernel, axis=0)

        # Construct the error
        ae = self._create_a(ye)
        return numpy.sum(a*self.kernel, axis=0), \
                    numpy.sqrt(numpy.sum(numpy.square(ae*self.kernel), axis=0))


def convolution_variable_sigma(y, sigma, ye=None, integral=False, large_sigma=10.):
    r"""
    Convolve a discretely sampled function :math:`y(x)` with a Gaussian
    kernel, :math:`g`, where the standard deviation of the kernel is a
    function of :math:`x`, :math:`\sigma(x)`.  Nominal calculations can
    be performed to propagate the error in the result; these
    calculations **do not** include the covariance between the pixels,
    which will mean that the calculations likely have significant error!

    The convolution is defined as:

    .. math::

        (y\ast g)(x) &= \int_{-\infty}^{\infty} y(X)\ g(\sigma,x-X)\ dX \\
                     &= \int_{-\infty}^{\infty} \frac{y(X)}{\sqrt{2\pi}\
                        \sigma(X)}\ \exp\left(-\frac{(x-X)^2}{2\
                        \sigma(X)^2}\right) dX .

    To minimize edge effects and account for the censoring of the data
    (finite range in :math:`x`), the convolution is actually calculated
    as a definite integral and normalized as follows:

    .. math::

        (y\ast g)(x) \sim\frac{
        \int_{x-n_\sigma*\sigma(x)}^{x+n_\sigma*\sigma(x)} y(X)\
        g(\sigma,x-X)\ dX}{
        \int_{x-n_\sigma*\sigma(x)}^{x+n_\sigma*\sigma(x)}
        g(\sigma,x-X)\ dX} .

    The above is identical to getting the weighted mean of :math:`y` at
    each position :math:`x`, where the weights are defined by a Gaussian
    kernel centered at :math:`x` with a variable dispersion.

    Use of this function requires:
        - *y* and *sigma* must be 1D vectors
        - *y* and *sigma* must be uniformly sampled on the same grid
        - *sigma* must be in pixel units.

    .. todo::
        This should be deprecated and/or merged with
        :class:`VariableGaussianKernel`.

    Args:
        y (`numpy.ndarray`_):
            A uniformly sampled function to convolve.
        sigma (`numpy.ndarray`_):
            The standard deviation of the Gaussian kernel sampled at
            the same positions as ``y``. The units of ``sigma``
            *must* be in pixels.
        ye (`numpy.ndarray`_, optional):
            Errors in the function :math:`y(x)`. Must be the same
            shape as ``y``.
        integral (:obj:`bool`, optional):
            Construct the kernel based on the integral of the
            Gaussian over the pixel width using the error function,
            instead of just sampling the Gaussian function directly.
            See :class:`VariableGaussianKernel`.
        large_sigma (:obj:`float`, optional):
            A convenience parameter that causes a warning to be
            issued if any of the values in ``sigma`` are larger than
            this threshold.

    Returns:
        `numpy.ndarray`_: Arrays with the convolved function
        :math:`(y\ast g)(x)` sampled at the same positions as the
        input :math:`x` vector and its error. The second array will
        be returned as None if the error vector is not provided.
    """
    if numpy.any(sigma > large_sigma):
        warnings.warn('Gaussian kernel dispersion larger than {0:.1f} pixels!'.format(large_sigma))
    return VariableGaussianKernel(sigma, integral=integral).convolve(y, ye=ye)


class SpectralResolution:
    r"""
    Container class for the resolution, :math:`R =
    \lambda/\Delta\lambda`, of a spectrum.  The primary functionality is
    to determine the parameters necessary to match the resolution of one
    spectrum to another.  It can also be used as a function to
    interpolate the spectral resolution at a given wavelenth.

    Args:
        wave (numpy.ndarray):
            A 1D vector with the wavelength in angstroms.  The sampling
            may be either in linear steps of wavelength or
            :math:`\log_{10}` steps.
        sres (numpy.ndarray):
            A 1D vector with the spectral resolution, :math:`R`, sampled
            at the positions of the provided wavelength vector.
        log10 (:obj:`bool`, optional):
            Flag that the spectrum has been binned logarithmically (base
            10) in wavelength
        interp_ext (:obj:`int`, :obj:`tuple`, :obj:`str`, optional):
            The value to pass as ``fill_value`` to the interpolator,
            which defines its behavior when attempting to sample the
            spectral resolution beyond where it is defined. See
            `scipy.interpolate.interp1d`_.
        bounds_error (:obj:`bool`, optional):
            Force `scipy.interpolate.interp1d`_ to raise an exception
            if the interpolation is outside the bounds of the defined
            spectral resolution function.

    Raises:
        ValueError: Raised if *wave* is not a 1D vector or if *wave* and
            *sres* do not have the same shape.

    Attributes:
        interpolator (`scipy.interpolate.interp1d`_):
            An object used to interpolate the spectral resolution at any
            given wavelength.  The interpolation is hard-wired to be
            *linear* and its extrapolation behavior is defined by
            ``interp_ext``.  The wavelength and resolution vectors are
            held by this object for later reference if needed.
        log10 (:obj:`bool`):
            Flag that the spectrum has been binned logarithmically (base
            10) in wavelength
        c (:obj:`float`):
            The speed of light; provided by `astropy.constants`_.
        dv (:obj:`float`):
            The velocity step per pixel in km/s.  Defined using
            :func:`spectrum_velocity_scale` if :attr:`log10` is True;
            otherwise set to None.
        dw (:obj:`float`):
            The wavelength step per pixel in angstroms.  Defined as the
            wavelength step between the first two pixels if
            :attr:`log10` is False; otherwise set to None.
        min_sig (:obj:`float`):
            Minimum standard deviation allowed for the kernel used to
            match two spectral resolutions.  See
            :func:`GaussianKernelDifference`.
        sig_pd (`numpy.ndarray`):
            The standard deviation in pixels required to match the
            spectral resolution of this object to the resolution defined
            by a different :class:`SpectralResolution` object.  See
            :func:`GaussianKernelDifference`.
        sig_mask (`numpy.ndarray`):
            A *uint* vector used to identify measurements of
            :attr:`sig_pd` that should **not** be used to match the
            spectral resolution of this object to the resolution defined
            by a different :class:`SpectralResolution` object.  See
            :func:`GaussianKernelDifference`.
        sig_vo (:obj:`float`):
            A constant offset of the kernel standard deviation **in
            km/s** that has been applied to :attr:`sig_pd`.  See
            :func:`GaussianKernelDifference`.
        
    .. todo::

        - Allow it to determine if the binning is linear or geometric,
          then use the *log10* keyword to distinguish between natural
          log and :math:`log_{10}` binning.
        - Allow for more than one type of line-spread function?

    .. warning::

        The default behavior of the interpolator is to extrapolate the
        input spectral resolution vector when trying to sample from
        regions beyond where it is sampled.  Use ``interp_ext`` change
        this; see `scipy.interpolate.interp1d`_.

    """
    def __init__(self, wave, sres, log10=False, interp_ext='extrapolate', bounds_error=False):
        # Check the sizes
        if len(wave.shape) != 1:
            raise ValueError('wave must be a 1D array!')
        if wave.shape != sres.shape:
            raise ValueError('wave and sres must have the same shape!')

        # Use linear interpolation when sampling the spectral resolution
        self.interpolator = interpolate.interp1d(wave, sres, fill_value=interp_ext,
                                                 bounds_error=bounds_error)
        self.log10 = log10
        self.c = astropy.constants.c.to('km/s').value

        self.dv = spectrum_velocity_scale(wave) if log10 else None
        self.dw = None if log10 else wave[1] - wave[0]

        # No resolution matching calculated yet
        self.min_sig = None
        self.sig_pd = None
        self.sig_mask = None
        self.sig_vo = None


    def __call__(self, w):
        """Interpolate the spectral resolution at wavelength *w*."""
        return self.interpolator(w)


    def _finalize_GaussianKernelDifference(self, sig2_pd):
        r"""
        Given the calculated :math:`\sigma^2_{p,d}`, calculate and save
        the attributes :attr:`sig_pd` and :attr:`sig_mask`.  See
        :func:`GaussianKernelDifference`.
        """
#        print('finalize b')
        indx = numpy.isclose(sig2_pd, 0.0)
        nindx = numpy.invert(indx)
        self.sig_pd = sig2_pd.copy()
        self.sig_pd[nindx] = sig2_pd[nindx]/numpy.sqrt(numpy.absolute(sig2_pd[nindx]))
        self.sig_pd[indx] = 0.0
#        self.sig_mask = numpy.array(self.sig_pd < -self.min_sig).astype(numpy.uint)
        self.sig_mask = numpy.array(self.sig_pd < self.min_sig).astype(numpy.uint)
#        print('finalize e')


    def _convert_vd2pd(self, sig2_vd):
        r"""
        Convert from :math:`\sigma^2_{v,d}` to :math:`\sigma^2_{p,d}`.
        See :func:`GaussianKernelDifference`.
        """
#        print('vd2pd')
        return sig2_vd / numpy.square(self.dv) if self.log10 else \
               sig2_vd / numpy.square(self.c*self.dw/self.wave())


    def _convert_pd2vd(self, sig2_pd):
        r"""
        Convert from :math:`\sigma^2_{p,d}` to :math:`\sigma^2_{v,d}`.
        See :func:`GaussianKernelDifference`.
        """
#        print('pd2vd')
        return sig2_pd * numpy.square(self.dv) if self.log10 else \
               sig2_pd * numpy.square(self.c*self.dw/self.wave())


    def wave(self, copy=False):
        """
        Return the wavelength vector; held by :attr:`interpolator`.
        """
#        return self.interpolator._data[0]
        return self.interpolator.x.copy() if copy else self.interpolator.x


    def sres(self, copy=False):
        """
        Return the resolution vector; held by :attr:`interpolator`.
        """
#        return self.interpolator._data[1]
        return self.interpolator.y.copy() if copy else self.interpolator.y


    def instrumental_dispersion(self, w=None):
        r"""

        Return the instrumental dispersion by converting from :math:`R`
        to :math:`\sigma_{v,inst}` according to:
            
        .. math::
            
            \sigma_{v,inst} = \frac{c}{\sqrt{8\ln 2}\ R}.

        If w is None, just convert the internal interpolator values.
        Otherwise, return the values sampled at w.
        """
        if w is None:
            return numpy.ma.divide(self.c, DAPConstants.sig2fwhm*self.interpolator.y).filled(0.0)
        return numpy.ma.divide(self.c, DAPConstants.sig2fwhm*self.interpolator(w)).filled(0.0)


    def match(self, new_sres, no_offset=True, min_sig_pix=0.0):
        """
        Currently only an alias for :func:`GaussianKernelDifference`.
        """
        self.GaussianKernelDifference(new_sres, no_offset=no_offset, min_sig_pix=min_sig_pix)


    def GaussianKernelDifference(self, new_sres, no_offset=True, min_sig_pix=0.0):
        r"""
        Determine the parameters of a Gaussian kernel required to
        convert the resolution of this object to the resolution of a
        different the :class:`SpectralResolution` object, *new_sres*.

        The spectral resolution is defined as :math:`R =
        \lambda/\Delta\lambda`, where :math:`\Delta\lambda` is the FWHM
        of the spectral resolution element.  The standard deviation of
        the resolution element in angstroms is then
   
        .. math::
    
            \sigma_\lambda = \frac{\lambda}{f R}, \ \ {\rm where} \ \  f
            = \frac{{\rm FWHM_\lambda}}{\sigma_\lambda} = \sqrt{8\ln 2}.

        Assuming a Gaussian (in angstroms) line-spread function:

        .. math::

            \sigma^2_{\lambda,2} = \sigma^2_{\lambda,1} +
            \sigma^2_{\lambda,d}

        such that

        .. math::

            \sigma^2_{\lambda,d} = \left(\frac{\lambda}{f}\right)^2
            (R^{-2}_2 - R^{-2}_1)

        is the defining parameter of the Gaussian kernel needed to take
        a spectrum of resolution :math:`R_1` to one with a resolution of
        :math:`R_2`.

        For input to :func:`convolution_variable_sigma`, the
        *wavelength-dependent* parameter of the Gaussian kernel is
        converted to pixels.  This function allows for the input spectra
        to be linearly sampled in angstroms or log10(angstroms).  For
        the former (*log10=False*), 

        .. math::

            \sigma^2_{p,d} = \left(\frac{\lambda}{f\
            \delta\lambda}\right)^2 (R^{-2}_2 - R^{-2}_1)

        where :math:`\delta\lambda` is the size of the pixel in
        angstroms.  If the units are log10(angstrom) (*log10=True*), we
        approximate the velocity width of each pixel to be :math:`\delta
        v = c \ln(10.0) (\log\lambda[1]-\log\lambda[0])`, such that
    
        .. math::

            \sigma^2_{p,d} &= \left(\frac{c}{ \delta v \lambda}\right)^2
            \sigma^2_{\lambda,d} \\ &= \left(\frac{c}{ f\ \delta
            v}\right)^2 (R^{-2}_2 - R^{-2}_1)\ ;

        :math:`c` is the speed of light in km/s.

        The nominal use of this algorithm assumes :math:`R_1 \geq R_2`.
        However, in practice, :func:`convolution_variable_sigma` only
        uses a Gaussian kernel up to some minimum value of
        :math:`\epsilon_\sigma`; below this, the kernel is assumed to be
        a Delta function.  Therefore, as long as
    
        .. math::
    
            \sigma_{p,d} \equiv \sigma^2_{p,d}/\sqrt{|\sigma^2_{p,d}|}
            \geq -\epsilon_\sigma\ ,
        
        the behavior of :func:`convolution_variable_sigma` should not be
        affected.  However, in regions with
    
        .. math::
    
            \sigma_{p,d} \equiv \sigma^2_{p,d}/\sqrt{|\sigma^2_{p,d}|}
            \leq \epsilon_\sigma\ ,

        the behavior of :func:`convolution_variable_sigma` does *not*
        produce an accurate convolution!        
    
        To deal with spectral regions that do not have
        :math:`\sigma_{p,d} \geq \epsilon_\sigma`, there are three
        choices:

            (**Option 1**) trim the spectral range to only those
            spectral regions where the existing resolution is better
            than the target resolution up to this limit,
        
            (**Option 2**) match the existing resolution to the target
            resolution up to some constant offset that must be accounted
            for in subsequent analyses, or

            (**Option 3**) allow for a wavelength dependent difference
            in the spectral resolution that must be accounted for in
            subsequent analyses.

        The choice of either Option 1 or 2 is selected by setting
        *no_offset* to, respectively, True or False; Option 1 is the
        default behavior.  Currently, Option 3 is not allowed.

        For Option 1, pixels with :math:`\sigma_{p,d} <
        \epsilon_\sigma` are masked (*sigma_mask = 1*); however, the
        returned values of :math:`\sigma_{p,d}` are left unchanged.

        For Option 2, we define

        .. math::

            \sigma^2_{v,o} = | {\rm min}(0.0, {\rm min}(\sigma^2_{v,d})
            - {\rm max}(\epsilon_\sigma \delta v)^2) |

        where :math:`\delta v` is constant for the logarithmically
        binned spectrum and is wavelength dependent for the linearly
        binned spectra; in the latter case, the velocity step is
        determined for each pixel::

            _wave = self.wave()
            dv = self.c * (2.0*(_wave[1:] - _wave[0:-1]) / (_wave[1:] + _wave[0:-1]))

        If :math:`\sigma^2_{v,o} > 0.0`, it must be that :math:`{\rm
        min}(\sigma^2_{v,d}) < {\rm max}(\epsilon_\sigma \delta v)^2`,
        such that an offset should be applied.  In that case, the
        returned kernel parameters are

        .. math::

            \sigma^\prime_{v,d} = \sqrt{\sigma^2_{v,d} +
            \sigma^2_{v,o}}\ .

        with the units converted to pixels using the equations above, no
        pixels are masked, and :math:`\sqrt{\sigma^2_{v,o}}` is returned
        for the offset.  Otherwise, the offset is set to 0.

        Args:
            new_sres (:class:`SpectralResolution`):
                Spectral resolution to match to.
            no_offset (:obj:`bool`, optional):
                Force :math:`\sigma^2_{v,o} = 0` by masking regions with
                :math:`\sigma_{p,d} < \epsilon_\sigma`; i.e., the value
                of this arguments selects Option 1 (True) or Option 2
                (False).
            min_sig_pix (:obj:`float`, optional):
                Minimum value of the standard deviation allowed before
                assuming the kernel is a Delta function.

        """
#        print('kernel difference')
        # Save the minimum pixel sigma to allow
        self.min_sig = min_sig_pix

        # Interpolate the new spectral resolution vector at the wavelengths
        # of the input spectral resolution
        _wave = self.wave()
        _sres = self.sres()
        interp_sres = new_sres(_wave)

        # Determine the variance (in angstroms) of Gaussian needed to match
        # input resolution to the new values
#        print('sig2_wd')
        sig2_wd = numpy.square(_wave/DAPConstants.sig2fwhm) \
                  * (1.0/numpy.square(interp_sres) - 1.0/numpy.square(_sres))

#        pyplot.plot(_wave, _sres)
#        pyplot.plot(_wave, interp_sres)
#        pyplot.show()

#        print('sig2_vd')
        # Convert to km/s
        sig2_vd = numpy.square(self.c/_wave) * sig2_wd

        # Option 1:
        if no_offset:
            # Convert the variance to pixel coordinates
            sig2_pd = sig2_vd / numpy.square(self.dv) if self.log10 else \
                      sig2_wd / numpy.square(self.dw)
            # No offset applied
            self.sig_vo = 0.0

        # Option 2:
        else:
            # 1% fudge so pixel at min_sig is not masked!
            fudge = 1.01
            
            # Calculate the velocity step of each pixel
            dv = self.c * (2.0*(_wave[1:] - _wave[0:-1]) / (_wave[1:] + _wave[0:-1]))
            # Get the needed *velocity* offset (this is the square)
            self.sig_vo = fudge*numpy.absolute(min(0.0, numpy.amin(sig2_vd)
                                                  - numpy.square(self.min_sig * numpy.amax(dv))))
            # Apply it if it's larger than 0
            if self.sig_vo > 0:
                sig2_vd += self.sig_vo
                self.sig_vo = numpy.sqrt(self.sig_vo)

            # Convert the variance to pixel coordinates
            sig2_pd = self._convert_vd2pd(sig2_vd)

        self._finalize_GaussianKernelDifference(sig2_pd)

#        print('kernel difference done')

#    def ZeroGaussianKernelDifference(self, min_sig_pix=0.0):
#        self.min_sig = min_sig_pix
#        sig2_pd = numpy.zeros(self.wave().shape, dtype=numpy.float64)
#        self._finalize_GaussianKernelDifference(sig2_pd)


    def offset_GaussianKernelDifference(self, offset):
        r"""
        If the properties required to match the resolution of one
        spectrum to another has already been calculated (see
        :func:`GaussianKernelDifference`), this allows for one to apply
        an additional offset.  The additional offset **must** be in km/s
        (not pixels).

        The offset is applied in quadrature; however, the offset can be
        negative such that one can reduce :attr:`sig_pd`.  Once
        converted to km/s, the offset is applied by calculating:

        .. math::
        
            \sigma^{\prime\ 2}_{v,d} = \sigma^{2}_{v,d} +
            \sigma_{off}|\sigma_{off}|\ .

        :attr:`sig_vo` is adjusted in the same way, and the change in
        :math:`\sigma^{\prime\ 2}_{v,d}` is then propagated to
        :attr:`sig_pd` and :attr:`sig_mask`.
        
        Args:
            offset (float): Value of the standard deviation in km/s to
                add in quadrature to a previously calculated
                :attr:`sig_pd`.
        
        Raises:
            ValueError: Raised if the kernel properties have not yet
                been defined.
        """
        if self.min_sig is None or self.sig_pd is None or self.sig_mask is None \
                or self.sig_vo is None:
#            print('WARNING: No kernel difference yet defined.  Assuming 0.')
#            self.ZeroGaussianKernelDifference()
            raise ValueError('No kernel defined yet.  Run GaussianKernelDifference first.')
        if numpy.isclose(offset,0.0):
            return
        off2 = offset*numpy.absolute(offset)
        sig2_vo = self.sig_vo*numpy.absolute(self.sig_vo) + off2
        self.sig_vo = 0.0 if numpy.isclose(sig2_vo, 0.0) \
                          else sig2_vo/numpy.sqrt(numpy.absolute(sig2_vo))
        sig2_vd = self._convert_pd2vd(self.sig_pd*numpy.absolute(sig_pd)) + off2
        self._finalize_GaussianKernelDifference(self._convert_vd2pd(sig2_vd))


    def adjusted_resolution(self, indx=None):
        r"""

        Return the resolution that should result from applying
        :func:`convolution_variable_sigma` to the spectrum associated
        with this spectral resolution object using :attr:`sigma_pd`.
        I.e., calculate:

        .. math::

            R_2 = \left[ \left(\frac{f}{c}\right)^2 \sigma^2_{v,d} +
            R^{-2}_1\right]^{-1/2}\ . 

        Args:
            indx (tuple): (**Optional**) Selection tuple used to return
                a subset of the full resolution vector.

        Returns:
            numpy.ndarray: The (full or selected) vector with the
            adjusted resolution.

        .. todo::
            Allow to reset the resolution of this object to the adjusted
            resolution and reset the kernel variables to None.

        """
        if indx is None:
            return 1.0/numpy.sqrt( numpy.square(DAPConstants.sig2fwhm/self.c) \
                                   * self._convert_pd2vd(self.sig_pd*numpy.absolute(self.sig_pd)) \
                                   + 1.0/numpy.square(self.sres()) )

        return 1.0/numpy.sqrt( numpy.square(DAPConstants.sig2fwhm/self.c) \
                            * (self._convert_pd2vd(self.sig_pd*numpy.absolute(self.sig_pd)))[indx] \
                               + 1.0/numpy.square(self.sres()[indx]) )


def match_spectral_resolution(wave, flux, sres, new_sres_wave, new_sres, ivar=None, mask=None, 
                              min_sig_pix=0.0, no_offset=True, variable_offset=False, log10=False,
                              new_log10=False, integral=True, quiet=False):
    r"""
    Adjust the existing spectral resolution of a spectrum to a **lower**
    resolution as best as possible.  The primary functionality is in
    :class:`SpectralResolution`, which determines the Gaussian kernel
    parameters needed to match the resolution, and
    :func:`convolve_variable_sigma`, which actually performs the
    convolution to match the resolution.

    In particular, see
    :func:`SpectralResolution.GaussianKernelDifference` for a
    description of how the kernel parameters are determined and how
    regions where the target resolution is **higher** than the existing
    resolution.  In this case, one of the options is to adopt an offset
    of the resolution (in km/s) that could be corrected for in
    subsequent analysis.  In this case, setting *variable_offset* to
    True allows the offset to be different for all input spectra.  If
    one expects to combine the spectra, the default behavior should be
    used, forcing all the spectra to have a constant offset.

    Args:
        wave (numpy.ndarray): A 1D or 2D (:math:`N_{\rm spec}\times
            N_{\rm pix}`) array with the wavelength in angstroms for a
            set of spectra.  The sampling may be either in linear steps
            of wavelength or :math:`\log_{10}` steps, as set using
            *log10*.
        flux (numpy.ndarray): A 1D or 2D (:math:`N_{\rm spec}\times
            N_{\rm pix}`) array with the flux sampled at the provided
            wavelengths.
        sres (numpy.ndarray): A 1D or 2D (:math:`N_{\rm spec}\times
            N_{\rm pix}`) array with the spectral resolution, :math:`R`,
            at the provided wavelengths.
        new_sres_wave (numpy.ndarray): A 1D vector with the wavelength
            in angstroms at which the new resolution of the input
            spectra has been sampled.  The sampling may be either in
            linear steps of wavelength or :math:`\log_{10}` steps, as
            set using *new_log10*.
        new_sres (numpy.ndarray): A 1D vector with the new resolution
            for the input spectra.
        ivar (numpy.ndarray): (**Optional**) A 1D or 2D (:math:`N_{\rm
            spec}\times N_{\rm pix}`) array with the inverse variance of
            the flux sampled at the provided wavelengths.  This vector
            is used to estimate the noise in the resolution-matched
            spectra.

            .. warning::
                The accuracy of the errors still remain untested!
            
        mask (numpy.ndarray): (**Optional**) A 1D or 2D (:math:`N_{\rm
            spec}\times N_{\rm pix}`) array with a *uint* mask for the
            flux sampled at the provided wavelengths.
        no_offset (bool): (**Optional**) Force :math:`\sigma^2_{v,o} =
            0` by masking regions with :math:`\sigma_{p,d} <
            -\epsilon_\sigma`; i.e., the value of this arguments selects
            Option 1 (True) or Option 2 (False).  See
            :func:`SpectralResolution.GaussianKernelDifference`.
        min_sig_pix (float): (**Optional**) Minimum value of the
            standard deviation in pixels allowed before assuming the
            kernel is a Delta function.
        variable_offset (bool): (**Optional**) Flag to allow the offset
            applied to each spectrum (when the input contains more than
            one spectraum) to be tailored to each spectrum.  Otherwise
            (*variable_offset=False*) the offset is forced to be the
            same for all spectra.
        log10 (bool): (**Optional**) Flag that the spectrum has been
            binned logarithmically (base 10) in wavelength
        new_log10 (bool): (**Optional**) Flag that the coordinates of
            the new spectral resolution are  spectrum as been binned
            logarithmically (base 10) in wavelength.

    Returns: 
        numpy.ndarray: Five objects are returned:

            - A 1D or 2D (:math:`N_{\rm spec}\times N_{\rm pix}`) array
              with the resolution-matched flux sampled at the input
              wavelengths.
            - A 1D or 2D (:math:`N_{\rm spec}\times N_{\rm pix}`) array
              with the spectral resolution, :math:`R`, of the
              resolution-matched spectra at the provided wavelengths.
            - A 1D vector with any constant offset in resolution **in
              km/s** between the targetted value and the result.  See
              :func:`SpectralResolution.GaussianKernelDifference`.
            - A 1D or 2D (:math:`N_{\rm spec}\times N_{\rm pix}`) array
              with a *uint* mask for the resolution-matched flux sampled
              at the input wavelengths.  This is returned regardless of
              whether an input mask was provided.  Any pixel that had a
              resolution that was lower than the target resolution (up
              to some tolerance defined by *min_sig_pix*) is returned as
              masked.
            - A 1D or 2D (:math:`N_{\rm spec}\times N_{\rm pix}`) array
              with the inverse variance of the resolution-matched flux
              sampled at the input wavelengths.  If *ivar*
              is not provided, a 'None' returned as the last element

    Raises:
        ValueError: Raised if:

            - the input *wave* array is 2D and the *sres* array is not;
              a 1D wavelength array is allowed for a 2D *sres* array but
              not vice versa

            - the number of spectral pixels in *wave*, *flux*, and
              *sres* is not the same

            - the shape of the *flux*, *mask* (if provided), and *ivar*
              (if provided) are not the same

            - the shape of the *new_sres_wave* and *new_sres* arrays
              are not the same and/or not 1D

    .. todo::

        - Add interp_ext != 'extrapolate' option?
        - Better way to use warnings?

    """
    # Check the dimensionality of wave and sres
    wave_matrix = len(wave.shape) == 2
    sres_matrix = len(sres.shape) == 2
    if wave_matrix and not sres_matrix:
        raise ValueError('If input wavelength array is 2D, the spectral resolution array must' \
                         ' also be 2D')

    # Check the shapes
    if (wave_matrix == sres_matrix and wave.shape != sres.shape) or \
       (not wave_matrix and sres_matrix and wave.shape[0] != sres.shape[1]):
        raise ValueError('Input spectral resolution and coordinate arrays must have the same' \
                         ' number of spectral channels!')
    if (wave_matrix and wave.shape != flux.shape) or \
       (not wave_matrix and len(flux.shape) == 2 and wave.shape[0] != flux.shape[1]) or \
       (not wave_matrix and len(flux.shape) == 1 and wave.shape != flux.shape):
        raise ValueError('Input flux and coordinate arrays must have the same number of' \
                         ' spectral channels!')
    if (mask is not None and mask.shape != flux.shape):
        raise ValueError('Input flux and mask arrays must have the same shape!')
    if (ivar is not None and ivar.shape != flux.shape):
        raise ValueError('Input flux and ivar arrays must have the same shape!')

    if len(sres.shape) > len(flux.shape):
        raise ValueError('Shape of the spectral resolution array must be <= to the flux array.')
        
    if len(new_sres_wave.shape) != 1 or len(new_sres.shape) != 1:
        raise ValueError('New spectral resolution and coordinate arrays must be 1D!')
    if new_sres_wave.shape != new_sres.shape:
        raise ValueError('New spectral resolution and coordinate arrays must have the same shape!')

    # Raise a warning if the new_sres vector will have to be
    # extrapolated for the input wavelengths
    if not quiet and (numpy.amin(wave) < new_sres_wave[0] or numpy.amax(wave) > new_sres_wave[-1]):
        warnings.warn('Mapping to the new spectral resolution will require extrapolating the' \
                      ' provided input vectors!')

    # Initialize some variables
    nspec = 1 if len(flux.shape) == 1 else flux.shape[0]
    nsres = 1 if len(sres.shape) == 1 else sres.shape[0]
    if sres_matrix and nspec != nsres:
        raise ValueError('For 2D matrices, number of spectral resolution vectors must match the ' \
                         'number of spectra.')
    spec_dim = len(flux.shape)
    sres_dim = len(sres.shape)
    sigma_offset = numpy.zeros(nspec, dtype=numpy.float64)
    new_res = SpectralResolution(new_sres_wave, new_sres, log10=new_log10,
                                 interp_ext=(new_sres[0],new_sres[-1]))

#    pyplot.plot(new_sres_wave, new_sres)
#    pyplot.show()
#    exit()

    res = numpy.empty(nspec, dtype=object)

    # Get the kernel parameters necessary to match all spectra to the
    # new resolution
    if nsres == 1 and sres_dim == 1:
        res[0] = SpectralResolution(wave, sres, log10=log10, interp_ext=(sres[0], sres[-1]))
        res[0].match(new_res, no_offset=no_offset, min_sig_pix=min_sig_pix)
        sigma_offset[0] = res[0].sig_vo
        for i in range(1,nspec):
            res[i] = res[0]
#        pyplot.plot(wave, res[0].sig_pd)
#        pyplot.show()
    else:
        for i in range(0,nsres):
            _wave = wave[i,:].ravel() if wave_matrix else wave
            _sres = sres[i,:].ravel() if sres_matrix else sres
            res[i] = SpectralResolution(_wave, _sres, log10=log10,
                                        interp_ext=(sres[i,0],sres[i,-1]))
            res[i].match(new_res, no_offset=no_offset, min_sig_pix=min_sig_pix)
            sigma_offset[i] = res[i].sig_vo
#            pyplot.plot(_wave, res[i].sig_pd)
#            pyplot.plot(_wave, numpy.sqrt(res[i]._convert_pd2vd(numpy.square(res[i].sig_pd))))
#            pyplot.show()

    # Force all the offsets to be the same, if requested
    if not no_offset and not variable_offset:
        common_offset = numpy.max(sigma_offset)
        offset_diff = numpy.sqrt( numpy.square(common_offset) - numpy.square(sigma_offset))
        for r, o in zip(res,offset_diff):
            r.offset_GaussianKernelDifference(o)

    # Perform the convolutions
    out_flux = flux.copy()
    out_ivar = None if ivar is None else numpy.ma.MaskedArray(ivar.copy())
    noise = None if ivar is None else numpy.ma.sqrt(1.0/out_ivar)
    out_sres = sres.copy()
    _mask = numpy.zeros(flux.shape, dtype=numpy.uint) if mask is None else mask
    out_mask = _mask.copy()

#    print('test div by zero')
    if nspec == 1 and spec_dim == 1:
        sig_kernel = res[0].sig_pd.copy()
        sig_kernel[sig_kernel < min_sig_pix] = 0.0
        indx = res[0].sig_pd > min_sig_pix
        try:
            if ivar is None:
                out_flux[indx] = convolution_variable_sigma(flux, sig_kernel,
                                                            integral=integral)[indx]
            else:
                oflux, oivar = convolution_variable_sigma(flux, sig_kernel,
                                                          ye=None if ivar is None else noise[indx],
                                                          integral=integral)

                out_flux[indx] = oflux[indx]
                out_ivar[indx] = oivar[indx]
                del oflux, oivar
            out_sres[indx] = res[0].adjusted_resolution(indx=indx)
            out_mask = numpy.array((res[0].sig_mask == 1) | (_mask == 1)).astype(numpy.uint)
        except ValueError as e:
            warnings.warn('Encountered ValueError: {0} ; continuing but resolution is NOT '
                          'changed and mask is set.'.format(e))
            out_mask = numpy.ones(flux.shape, dtype=numpy.uint)
    else:
        for i in range(0,nspec):
            if not quiet:
                print('Matching resolution: {0}/{1}'.format(i+1,nspec), end='\r')
            try:
#                indx = numpy.where(res[i].sig_pd > min_sig_pix)
                sig_kernel = res[i].sig_pd.copy()
                sig_kernel[sig_kernel < min_sig_pix] = 0.0
                indx = res[i].sig_pd > min_sig_pix
                if ivar is None:
                    out_flux[i,indx] = convolution_variable_sigma(flux[i,:], sig_kernel,
                                                                  integral=integral)[indx]
                else:
                    oflux, oivar = convolution_variable_sigma(flux[i,:], sig_kernel,
                                                              ye=None if ivar is None 
                                                                    else noise[i,:].ravel(),
                                                              integral=integral)
                    out_flux[i,indx] = oflux[indx]
                    out_ivar[i,indx] = oivar[indx]
                    del oflux, oivar
                out_mask[i,:] = numpy.array((res[i].sig_mask == 1) \
                                        | (_mask[i,:] == 1)).astype(numpy.uint)
                if len(out_sres.shape) == 1 and i == 0:
                    out_sres[indx] = res[i].adjusted_resolution(indx=indx)
                else:
                    out_sres[i,indx] = res[i].adjusted_resolution(indx=indx)
            except ValueError as e:
                warnings.warn('Encountered ValueError: {0} ; continuing but resolution is NOT '
                              'changed and mask is set.'.format(e))
                out_mask[i,:] = numpy.ones(flux.shape[1], dtype=numpy.uint)
        if not quiet:
            print('Matching resolution: {0}/{0}'.format(nspec))

    # TODO: Add this functionality from the IDL version?
    #
    # Finally, the code masks a number of pixels at the beginning and
    # end of the spectra to remove regions affected by errors in the
    # convolution due to the censoring of the data.  The number of
    # pixels is the FWHM of the largest Gaussian applied in the
    # convolution: ceil(sig2fwhm*max(diff_sig_w)/dw).  This is currently
    # hard-wired and should be tested.

    if ivar is not None:
        out_ivar = numpy.square(1.0/out_ivar)
        # When returning out_ivar, convert it to a normal array
        return out_flux, out_sres, sigma_offset, out_mask, numpy.asarray(out_ivar)
    return out_flux, out_sres, sigma_offset, out_mask, None
    




from IPython import embed

import numpy

from mangadap.util.vander import Legendre1D


def test_1d_fit():
    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    y = 3 + 0.2*x + x*x - x*x*x 
    y += rng.normal(scale=0.01, size=x.size)

    c = numpy.polynomial.legendre.legfit(x,y,4)
    leg = Legendre1D(x,4)
    _c = leg.fit(y)

    assert numpy.allclose(c, _c), 'Legendre1D should match numpy.'


def test_2d_fit():
    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    c = numpy.array([[3, 0.2,   1,   -1],
                     [4, 0.4, 0.8, -0.8]])

    y = numpy.sum([_c[:,None] * numpy.power(x,i)[None,:] for i,_c in enumerate(c.T)], axis=0)
    y += rng.normal(scale=0.01, size=y.size).reshape(y.shape)

    c = numpy.polynomial.legendre.legfit(x,y.T,4)

    leg = Legendre1D(x,4)
    _c = leg.fit(y)

    assert numpy.allclose(c, _c.T), 'Legendre1D should match numpy.'


def test_w1d_1d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    y = 3 + 0.2*x + x*x - x*x*x 
    # Add noise
    y += rng.normal(scale=0.01, size=x.size)
    # Add in deviates
    indx = numpy.unique(rng.integers(1000, size=10))
    y[indx] += rng.normal(scale=10, size=len(indx))
    w = numpy.ones(y.shape, dtype=float)
    w[indx] = 0.

    c_nw = numpy.polynomial.legendre.legfit(x,y,4)
    c = numpy.polynomial.legendre.legfit(x,y,4,w=w)

    leg = Legendre1D(x,4)
    _c_nw = leg.fit(y)
    _c = leg.fit(y,w=w)

    assert numpy.allclose(c_nw, _c_nw), 'Legendre1D fits should match numpy.'
    assert numpy.allclose(c, _c), 'Weighted Legendre1D fits should match numpy.'


def test_w1d_2d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    c = numpy.array([[3, 0.2,   1,   -1],
                     [4, 0.4, 0.8, -0.8]])

    y = numpy.sum([_c[:,None] * numpy.power(x,i)[None,:] for i,_c in enumerate(c.T)], axis=0)
    # Add noise
    y += rng.normal(scale=0.01, size=y.size).reshape(y.shape)
    # Add in deviates
    indx = numpy.unique(rng.integers(1000, size=10))
    y[0,indx] += rng.normal(scale=10, size=len(indx))
    y[1,indx] += rng.normal(scale=10, size=len(indx))
    # Weight all y vectors the same
    w = numpy.ones(x.shape, dtype=float)
    w[indx] = 0.

    c_nw = numpy.polynomial.legendre.legfit(x,y.T,4)
    c = numpy.polynomial.legendre.legfit(x,y.T,4,w=w)

    leg = Legendre1D(x,4)
    _c_nw = leg.fit(y)
    _c = leg.fit(y,w=w)

    assert numpy.allclose(c_nw, _c_nw.T), 'Legendre1D fits should match numpy.'
    assert numpy.allclose(c, _c.T), 'Weighted Legendre1D fits should match numpy.'


def test_w2d_2d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    c = numpy.array([[3, 0.2,   1,   -1],
                     [4, 0.4, 0.8, -0.8]])

    y = numpy.sum([_c[:,None] * numpy.power(x,i)[None,:] for i,_c in enumerate(c.T)], axis=0)
    # Add noise
    y += rng.normal(scale=0.01, size=y.size).reshape(y.shape)
    # Add in deviates
    indx = numpy.unravel_index(numpy.unique(rng.integers(y.size, size=20)), y.shape)
    y[indx] += rng.normal(scale=10, size=len(indx[0]))
    # Weight y vectors independently
    w = numpy.ones(y.shape, dtype=float)
    w[indx] = 0.

    c = numpy.zeros((y.shape[0],5), dtype=float)
    for i in range(y.shape[0]):
        c[i] = numpy.polynomial.legendre.legfit(x,y[i],4,w=w[i])

    leg = Legendre1D(x,4)
    _c = leg.fit(y,w=w)

    assert numpy.allclose(c, _c), 'Weighted Legendre1D fits should match numpy.'

    # Tests both the construction of the model and that the residuals
    # are close to the expectation from the noise added to the data
    # (i.e. is the fit good).
    resid = numpy.ma.std(numpy.ma.MaskedArray(y, mask=numpy.logical_not(w>0)) - leg.model(_c))
    assert numpy.isclose(numpy.round(resid, decimals=2), 0.01), 'Residuals too large'


def test_glbrej_1d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    y = 3 + 0.2*x + x*x - x*x*x 
    # Add noise
    sigma = 0.01
    y += rng.normal(scale=sigma, size=x.size)
    # Add in deviates
    outlier = numpy.unique(rng.integers(1000, size=10))
    offset = rng.normal(scale=10, size=len(outlier))
    indx = numpy.absolute(offset) > 10*sigma
    outlier = outlier[indx]
    y[outlier] += offset[indx] 

    leg = Legendre1D(x,4)
    c, rej = leg.fit(y, rej_iter=-1)

    # Rejected all the outliers?
    assert numpy.all(rej[outlier]), 'Should have rejected all the outliers'


def test_locrej_1d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    y = 3 + 0.2*x + x*x - x*x*x 
    # Add noise
    sigma = 0.01
    y += rng.normal(scale=sigma, size=x.size) * numpy.absolute(x) * 5
    # Add in deviates
    outlier = numpy.unique(rng.integers(1000, size=10))
    offset = rng.normal(scale=10, size=len(outlier))
    indx = numpy.absolute(offset) > 10*sigma
    outlier = outlier[indx]
    y[outlier] += offset[indx] 

    leg = Legendre1D(x,4)
    c, rej = leg.fit(y, rej_iter=-1, rej_win=51)

    # Rejected all the outliers?
    assert numpy.all(rej[outlier]), 'Should have rejected all the outliers'

    # Don't use a local sigma
    _c, _rej = leg.fit(y, rej_iter=-1)

    # Should reject more when not considering local sigma
    assert numpy.sum(_rej) > numpy.sum(rej), 'Should have rejected more when using global sigma'

    # Don't use a local sigma and only one iteration
    _c, __rej = leg.fit(y, rej_iter=1)

    # Should reject more when allowed to continue until no more rejections
    assert numpy.sum(_rej) > numpy.sum(__rej), 'Should have rejected fewer in one iteration'


def test_glbrej_2d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    c = numpy.array([[3, 0.2,   1,   -1],
                     [4, 0.4, 0.8, -0.8]])

    y = numpy.sum([_c[:,None] * numpy.power(x,i)[None,:] for i,_c in enumerate(c.T)], axis=0)
    # Add noise
    sigma = 0.01
    y += rng.normal(scale=sigma, size=y.size).reshape(y.shape)
    # Add in deviates
    outlier = numpy.unravel_index(numpy.unique(rng.integers(y.size, size=20)), y.shape)
    offset = rng.normal(scale=10, size=len(outlier[0]))
    indx = numpy.absolute(offset) > 10*sigma
    outlier = (outlier[0][indx], outlier[1][indx])
    y[outlier] += offset[indx] 

    leg = Legendre1D(x,4)
    c, rej = leg.fit(y, rej_iter=-1)

    # Rejected all the outliers?
    assert numpy.all(rej[outlier]), 'Should have rejected all the outliers'


def test_locrej_2d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    c = numpy.array([[3, 0.2,   1,   -1],
                     [4, 0.4, 0.8, -0.8]])

    y = numpy.sum([_c[:,None] * numpy.power(x,i)[None,:] for i,_c in enumerate(c.T)], axis=0)
    # Add noise
    sigma = 0.01
    y += rng.normal(scale=sigma, size=y.size).reshape(y.shape) \
            * numpy.absolute(x)[None,:] * 5
    # Add in deviates
    outlier = numpy.unravel_index(numpy.unique(rng.integers(y.size, size=20)), y.shape)
    offset = rng.normal(scale=10, size=len(outlier[0]))
    indx = numpy.absolute(offset) > 10*sigma
    outlier = (outlier[0][indx], outlier[1][indx])
    y[outlier] += offset[indx] 

    leg = Legendre1D(x,4)
    c, rej = leg.fit(y, rej_iter=-1, rej_win=51)

    # Rejected all the outliers?
    assert numpy.all(rej[outlier]), 'Should have rejected all the outliers'

    # Don't use a local sigma
    _c, _rej = leg.fit(y, rej_iter=-1)

    # Should reject more when not considering local sigma
    assert numpy.sum(_rej) > numpy.sum(rej), 'Should have rejected more when using global sigma'


def test_ew_glbrej_1d_fit():

    rng = numpy.random.default_rng(seed=8001)

    x = numpy.linspace(-1, 1, 1000)
    y = 3 + 0.2*x + x*x - x*x*x 

    # Add noise
    sigma = 0.01
    noise = rng.normal(scale=sigma, size=x.size)
    err = numpy.full(y.shape, sigma, dtype=float)

    outlier = numpy.unique(rng.integers(1000, size=10))
    offset = rng.normal(scale=10., size=len(outlier))
    noise[outlier] = offset
    err[outlier] = 10.

    # Add noise including larger error
    y += noise

    leg = Legendre1D(x,4)
    c, rej = leg.fit(y, err=err, rej_iter=-1)
    _c, _rej = leg.fit(y, rej_iter=-1)

    # Fewer points should be rejected when accounting for errors
    assert numpy.sum(_rej) > numpy.sum(rej), \
            'Should not have rejected as many points when accounting for error'






from IPython import embed

from mangadap.scripts import scriptbase


class RunDap(scriptbase.ScriptBase):
    """
    Simple wrapper class for the rundap script.
    """
    @classmethod
    def name(cls):
        """
        Return the name of the executable.
        """
        return 'rundap'

    @classmethod
    def get_parser(cls, width=None):

        parser = super().get_parser(description='Perform analysis of integral-field data.',
                                    width=width)

        # Read the optional run-mode arguments
        parser.add_argument("--overwrite",
                            help="if all selected, will run dap for all plates/ifudesigns/modes "
                            " regardless of state", action="store_true", default=False)
        parser.add_argument('-v', '--verbose', action='count',
                            help='Set verbosity level for manga_dap; can be omitted and set up '
                                 'to -vv', default=0)
        parser.add_argument("--quiet", help="suppress screen output", action="store_true",
                            default=False)
        parser.add_argument("--print_version", help="print DAP version and stop",
                            action="store_true", default=False)
       
        # These arguments are used to override default behavior
        parser.add_argument("--drpver", type=str, default=None,
                            help='MaNGA DRP version for analysis; $MANGADRP_VER by default')
        parser.add_argument("--redux_path", type=str, help="main DRP output path", default=None)
        parser.add_argument("--dapver", type=str, default=None,
                            help='optional output version, different from product version.  This '
                                 '*only* affects the output directory structure.  It does *not* '
                                 'select the version of the DAP to use.')
        parser.add_argument("--analysis_path", type=str, help="main DAP output path", default=None)

        parser.add_argument("--plan_file", type=str, help="parameter file with the MaNGA DAP "
                            "execution plan to use instead of the default" , default=None)
        parser.add_argument("--platelist", type=str, help="set list of plates to reduce",
                            default=None)
        parser.add_argument("--ifudesignlist", type=str, help="set list of ifus to reduce",
                            default=None)
        parser.add_argument("--list_file", type=str,
                            help='A file with the list of plates and ifudesigns to analyze',
                            default=None)
        parser.add_argument("--combinatorics", help="force execution of all permutations of the "
                            "provided lists", action="store_true", default=False)

        parser.add_argument('--sres_ext', type=str, default=None,
                            help='Spectral resolution extension to use.  Default set by '
                                 'MaNGADataCube class.')
        parser.add_argument('--sres_fill', type=str, default=None,
                            help='If present, use interpolation to fill any masked pixels in the '
                                 'spectral resolution vectors. Default set by MaNGADataCube '
                                 'class.')
        parser.add_argument('--covar_ext', type=str, default=None,
                            help='Use this extension to define the spatial correlation matrix.  '
                                 'Default set by MaNGADataCube class.')

        parser.add_argument('--on_disk', action='store_true', default=False,
                            help='When using the DRPall file to collate the data for input to '
                                 'the DAP, search for available DRP files on disk instead of '
                                 'using the DRPall file content.')
        parser.add_argument('--can_analyze', action='store_true', default=False,
                            help='Only construct script files for datacubes that can/should be '
                                 'analyzed by the DAP.  See '
                                 ':func:`~mangadap.survey.drpcomplete.DRPComplete.can_analyze`.')

        parser.add_argument("--log", help="Have the main DAP executable produce a log file",
                            action="store_true", default=False)
        parser.add_argument("--no_proc", help="Do NOT perform the main DAP processing steps",
                            action="store_true", default=False)
        parser.add_argument("--no_plots", help="Do NOT create QA plots", action="store_true",
                            default=False)
        parser.add_argument("--post", help="Create/Submit the post-processing scripts",
                            action="store_true", default=False)
        parser.add_argument("--post_plots", action="store_true", default=False,
                            help="Create/Submit the post-processing plotting scripts")

        # Read arguments specific to the cluster submission behavior
        parser.add_argument("--label", type=str, help='label for cluster job', default='mangadap')
        parser.add_argument("--nodes", type=int, help='number of nodes to use in cluster',
                            default=1)
        parser.add_argument("--cpus", type=int,
                            help='number of cpus to use per node.  Default is to use all available'
                                 '; otherwise, set to minimum of provided number and number of '
                                 'processors per node', default=None)
        parser.add_argument("--fast", dest='qos', type=str, help='qos state', default=None)
        parser.add_argument("--umask", type=str, help='umask bit for cluster job', default='0027')
        parser.add_argument("--walltime", type=str, help='walltime for cluster job',
                            default='240:00:00')
        parser.add_argument("--toughness", dest='hard', action='store_false', default=True,
                            help='turn off hard keyword for cluster submission')

        parser.add_argument('--create', action='store_true', default=False,
                            help='use the pbs package to create the cluster scripts')
        parser.add_argument('--submit', action='store_true', default=False,
                            help='submit the scripts to the cluster')
        parser.add_argument('--progress', action='store_true', default=False,
                            help='instead of closing the script, report the progress of the '
                                 'analysis on the cluster; this is required if you want to submit '
                                 'the DAPall script immediately after completing the individual '
                                 'cube analysis')

        parser.add_argument("--queue", dest='queue', type=str, help='set the destination queue',
                            default=None)
        return parser

    @staticmethod
    def main(args):

        if args.print_version:
            from mangadap import __version__
            print(f'DAP Version: {__version__}')
            return

        from mangadap.survey.rundap import rundap

        # Handle special argument cases
        if args.qos is not None and args.nodes > 1:
            warnings.warn('Requesting the fast node requires node=1.  Ignoring input node number.')
            nodes = 1               # Force the number of nodes to be 1
        else:
            nodes = args.nodes

        _rundap = rundap(overwrite=args.overwrite, quiet=args.quiet, drpver=args.drpver,
                         redux_path=args.redux_path, dapver=args.dapver,
                         analysis_path=args.analysis_path, plan_file=args.plan_file,
                         platelist=args.platelist, ifudesignlist=args.ifudesignlist,
                         combinatorics=args.combinatorics, list_file=args.list_file,
                         sres_ext=args.sres_ext, sres_fill=args.sres_fill,
                         covar_ext=args.covar_ext, on_disk=args.on_disk,
                         can_analyze=args.can_analyze, log=args.log, dapproc=not args.no_proc,
                         pltifu_plots=not args.no_plots, post_process=args.post,
                         post_plots=args.post_plots, report_progress=args.progress,
                         verbose=args.verbose, label=args.label, nodes=nodes, cpus=args.cpus,
                         qos=args.qos, umask=args.umask, walltime=args.walltime, hard=args.hard,
                         create=args.create, submit=args.submit, queue=args.queue)






#!/usr/bin/env python3

import os
import time
import numpy

from matplotlib import pyplot, ticker, rc

from astropy.io import fits
import astropy.constants

from mangadap.proc.util import growth_lim
from mangadap.util.bitmask import BitMask

#-----------------------------------------------------------------------------
def init_ax(fig, pos):
    ax = fig.add_axes(pos, facecolor='0.98')
    ax.minorticks_on()
    ax.tick_params(which='major', length=6) #, direction='in')
    ax.tick_params(which='minor', length=3) #, direction='in')
    return ax


def redshift_distribution(dapall):

    indx = (dapall['DAPDONE'] == 1) & (dapall['DAPTYPE'] == 'HYB10-GAU-MILESHC')
    nsa_indx = indx & (dapall['NSA_Z'] > -999)
    str_indx = indx & (dapall['STELLAR_Z'] > -999)
    gas_indx = indx & (dapall['HA_Z'] > -999)

    font = { 'size' : 14 }
    rc('font', **font)

    w,h = pyplot.figaspect(1)
    fig = pyplot.figure(figsize=(1.5*w,1.5*h))

    offset_lim = [-90, 90]
    ax = init_ax(fig, [0.17, 0.10, 0.56, 0.3])
    ax.set_xlim([-0.01, 0.16])
    ax.set_ylim(offset_lim)
    ax.grid(True, which='major', color='0.8', zorder=0, linestyle='-')

    dcz = astropy.constants.c.to('km/s').value * (dapall['STELLAR_Z'][str_indx & nsa_indx]
                                                  - dapall['NSA_Z'][str_indx & nsa_indx])
    ax.scatter(dapall['NSA_Z'][str_indx & nsa_indx], dcz,
                   marker='.', s=10, lw=0, color='k', zorder=3)
    ax.text(0.5, -0.2, r'$z_{\rm NSA}$', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes)
    ax.text(-0.18, 0.5, r'$c (z_\ast - z_{\rm NSA})$ [km/s]', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes, rotation='vertical')

    ax = init_ax(fig, [0.73, 0.10, 0.1, 0.3])
    ax.set_ylim(offset_lim)
    ax.tick_params(which='both', direction='in', bottom=False)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.xaxis.set_major_formatter(ticker.NullFormatter())
    ax.grid(True, which='major', axis='y', color='0.8', zorder=0, linestyle='-')
    ax.hist(dcz, range=offset_lim, bins=100, alpha=0.7, color='0.4', histtype='stepfilled',
            orientation='horizontal', zorder=3)
    ax.hist(dcz, range=offset_lim, bins=100, alpha=0.7, color='k', histtype='step', lw=0.5,
            orientation='horizontal', zorder=4)

    ax = init_ax(fig, [0.22, 0.50, 0.56, 0.3])
    ax.set_xlim(offset_lim)
    ax.tick_params(which='both', left=False)
    ax.yaxis.set_major_formatter(ticker.NullFormatter())
    ax.grid(True, which='major', axis='x', color='0.8', zorder=0, linestyle='-')
    dcz = astropy.constants.c.to('km/s').value * (dapall['STELLAR_Z'][gas_indx & str_indx]
                                                  - dapall['HA_Z'][gas_indx & str_indx]),
    ax.hist(dcz, range=offset_lim, bins=100, alpha=0.7, color='0.4', histtype='stepfilled',
            zorder=3)
    ax.hist(dcz, range=offset_lim, bins=100, alpha=0.7, color='k', histtype='step', lw=0.5,
            zorder=4)
    ax.text(0.5, -0.2, r'$c (z_\ast - z_{\rm gas})$ [km/s]', horizontalalignment='center',
             verticalalignment='center', transform=ax.transAxes)

    ofile = '../ms/figs/redshift_difference.pdf'
    fig.canvas.print_figure(ofile, bbox_inches='tight')
    fig.clear()
    pyplot.close(fig)

#-----------------------------------------------------------------------------

if __name__ == '__main__':
    t = time.clock()

    dapall_file = '/Users/westfall/Work/MaNGA/testing/dr15/dapall-v2_4_3-2.2.1.fits'
    dapall = fits.open(dapall_file)['DAPALL'].data
    redshift_distribution(dapall)

    print('Elapsed time: {0} seconds'.format(time.clock() - t))





from IPython import embed

import numpy
from astropy.io import fits

from mangadap.par.artifactdb import ArtifactDB
from mangadap.par.emissionlinedb import EmissionLineDB
from mangadap.util.pixelmask import SpectralPixelMask
from mangadap.tests.util import data_test_file


def test_pixelmask():
    specfile = data_test_file('MaNGA_test_spectra.fits.gz')
    hdu = fits.open(specfile)
    pixelmask = SpectralPixelMask(artdb=ArtifactDB.from_key('BADSKY'),
                                  emldb=EmissionLineDB.from_key('ELPSCMSK'))
    assert numpy.sum(pixelmask.boolean(hdu['WAVE'].data, nspec=1)) == 489, \
                'Incorrect number of masked pixels'
